import warnings

from collections.abc import Mapping
from sqlalchemy import sql
from siuba.dply.verbs import join, left_join, right_join, inner_join, semi_join, anti_join

from ..backend import LazyTbl
from ..utils import _sql_select


def _joined_cols(left_cols, right_cols, on_keys, how, suffix = ("_x", "_y")):
    """Return labeled columns, according to selection rules for joins.

    Rules:
        1. For join keys, keep left table's column
        2. When keys have the same labels, add suffix
    """

    # TODO: remove sets, so uses stable ordering
    # when left and right cols have same name, suffix with _x / _y
    keep_right = set(right_cols.keys()) - set(on_keys.values())
    shared_labs = set(left_cols.keys()).intersection(keep_right)

    right_cols_no_keys = {k: right_cols[k] for k in keep_right}

    # for an outer join, have key columns coalesce values

    left_cols = {**left_cols}
    if how == "full":
        for lk, rk in on_keys.items():
            col = sql.functions.coalesce(left_cols[lk], right_cols[rk])
            left_cols[lk] = col.label(lk)
    elif how == "right":
        for lk, rk in on_keys.items():
            # Make left key columns actually be right ones (which contain left + extra)
            left_cols[lk] = right_cols[rk].label(lk)


    # create labels ----
    l_labs = _relabeled_cols(left_cols, shared_labs, suffix[0])
    r_labs = _relabeled_cols(right_cols_no_keys, shared_labs, suffix[1])

    return l_labs + r_labs
    


def _relabeled_cols(columns, keys, suffix):
    # add a suffix to all columns with names in keys
    cols = []
    for k, v in columns.items():
        new_col = v.label(k + str(suffix)) if k in keys else v
        cols.append(new_col)
    return cols


@join.register(LazyTbl)
def _join(left, right, on = None, *args, by = None, how = "inner", sql_on = None):
    _raise_if_args(args)

    if on is None and by is not None:
        on = by

    # Needs to be on the table, not the select
    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on)
    how = _validate_join_arg_how(how)
    
    # for equality join used to combine keys into single column
    consolidate_keys = on if sql_on is None else {}
    
    if how == "right":
        # switch joins, since sqlalchemy doesn't have right join arg
        # see https://stackoverflow.com/q/11400307/1144523
        left_sel, right_sel = right_sel, left_sel
        on = {v:k for k,v in on.items()}

    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create join ----
    join = left_sel.join(
            right_sel,
            onclause = bool_clause,
            isouter = how != "inner",
            full = how == "full"
            )

    # if right join, set selects back
    if how == "right":
        left_sel, right_sel = right_sel, left_sel
        on = {v:k for k,v in on.items()}

    # note, shared_keys assumes on is a mapping...
    # TODO: shared_keys appears to be for when on is not specified, but was unused
    #shared_keys = [k for k,v in on.items() if k == v]
    labeled_cols = _joined_cols(
            left_sel.columns,
            right_sel.columns,
            on_keys = consolidate_keys,
            how = how
            )

    sel = _sql_select(labeled_cols).select_from(join)
    return left.append_op(sel, order_by = tuple())


@semi_join.register(LazyTbl)
def _semi_join(left, right = None, on = None, *args, by = None, sql_on = None):
    if on is None and by is not None:
        on = by

    _raise_if_args(args)

    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on, left_sel, right_sel)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    exists_clause = _sql_select([sql.literal(1)]) \
            .select_from(right_sel) \
            .where(bool_clause)

    # only keep left hand select's columns ----
    sel = _sql_select(left_sel.columns) \
            .select_from(left_sel) \
            .where(sql.exists(exists_clause))

    return left.append_op(sel, order_by = tuple())


@anti_join.register(LazyTbl)
def _anti_join(left, right = None, on = None, *args, by = None, sql_on = None):
    if on is None and by is not None:
        on = by

    _raise_if_args(args)

    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on, left, right)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    #not_exists = ~sql.exists([1], from_obj = right_sel).where(bool_clause)
    exists_clause = _sql_select([sql.literal(1)]) \
            .select_from(right_sel) \
            .where(bool_clause)

    sel = left_sel.select().where(~sql.exists(exists_clause))
    
    return left.append_op(sel, order_by = tuple())
       
def _raise_if_args(args):
    if len(args):
        raise NotImplemented("*args is reserved for future arguments (e.g. suffix)")

def _validate_join_arg_on(on, sql_on = None, lhs = None, rhs = None):
    # handle sql on case
    if sql_on is not None:
        if on is not None:
            raise ValueError("Cannot specify both on and sql_on")

        return sql_on

    # handle general cases
    if on is None:
        # TODO:  currently, we check for lhs and rhs tables to indicate whether
        #        a verb supports inferring columns. Otherwise, raise an error.
        if lhs is not None and rhs is not None:
            # TODO: consolidate with duplicate logic in pandas verb code
            warnings.warn(
                "No on column passed to join. "
                "Inferring join columns instead using shared column names."
            )

            on_cols = list(set(lhs.columns.keys()).intersection(set(rhs.columns.keys())))

            if not on_cols:
                raise ValueError(
                    "No join column specified, or shared column names in join."
                )

            # trivial dict mapping shared names to themselves
            warnings.warn("Detected shared columns: %s" % on_cols)
            on = dict(zip(on_cols, on_cols))

        else:
            raise NotImplementedError("on arg currently cannot be None (default) for SQL")
    elif isinstance(on, str):
        on = {on: on}
    elif isinstance(on, (list, tuple)):
        on = dict(zip(on, on))


    if not isinstance(on, Mapping):
        raise TypeError("on must be a Mapping (e.g. dict)")

    return on

def _validate_join_arg_how(how):
    how_options = ("inner", "left", "right", "full")
    if how not in how_options:
        raise ValueError("how argument needs to be one of %s" %how_options)
    
    return how

def _create_join_conds(left_sel, right_sel, on):
    left_cols  = left_sel.columns  #lift_inner_cols(left_sel)
    right_cols = right_sel.columns #lift_inner_cols(right_sel)

    if callable(on):
        # callable, like with sql_on arg
        conds = [on(left_cols, right_cols)]
    else:
        # dict-like of form {left: right}
        conds = []
        for l, r in on.items():
            col_expr = left_cols[l] == right_cols[r]
            conds.append(col_expr)
            
    return sql.and_(*conds)



# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg, extend_base, win_agg,
        SqlTranslator,
        annotate, sql_scalar, sql_agg, win_cumul, sql_not_impl,
        AggOver, RankOver, Over,
        )

import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql
from sqlalchemy.sql import func as fn
from . import _dt_generics as _dt

# Custom dispatching in call trees ============================================

class BigqueryColumn(SqlColumn): pass
class BigqueryColumnAgg(SqlColumnAgg, BigqueryColumn): pass

# Custom translations =========================================================

@annotate(result_type="float")
def sql_floordiv(_, x, y):
    return sql.func.floor(x / y)

# datetime ----

@_dt.date_trunc.register
def _date_trunc(_: BigqueryColumn, col, name):
    return fn.datetime_trunc(col, sql.text(name))


@_dt.sql_func_last_day_in_period.register
def sql_func_last_day_in_period(_: BigqueryColumn, col, period):
    return fn.last_day(col, sql.text(period))


@_dt.sql_is_last_day_of.register
def sql_is_last_day_of(codata: BigqueryColumn, col, period):
    last_day = sql_func_last_day_in_period(codata, col, period)
    return _date_trunc(codata, col, "DAY") == last_day


def sql_extract(field):
    return lambda _, col: fn.extract(field, col)


# string ----

def sql_str_replace(_, col, pat, repl, n=-1, case=None, flags=0, regex=True):
    if n != -1 or case is not None or flags != 0:
        raise NotImplementedError("only pat and repl arguments supported in sql")

    if regex:
        return fn.regexp_replace(col, pat, repl)

    return fn.replace(col, pat, repl)

def sql_str_contains(_, col, pat, case=None, flags=0, na=None, regex=True):
    if case is not None or flags != 0:
        raise NotImplementedError("only pat and repl arguments supported in sql")

    if regex:
        return fn.regexp_contains(col, pat)

    return col.contains(pat)

# error messages ----

QUANTILE_ERROR = ("taking the median or quantile with percentile_cont can only be done in a mutate, "
                 "since this function cannot be used as an aggregate")

# any / all ----

def sql_any(window = False):
    f_win = AggOver if window else lambda x: x
    
    @annotate(input_type="bool")
    def f(_, col):
        return f_win(fn.sum(fn.cast(col, sa_types.Integer()))) != 0

    return f

def sql_all(window = False):
    f_win = AggOver if window else lambda x: x
    
    @annotate(input_type="bool")
    def f(_, col):
        # similar to any, but uses (not cond) summed is 0
        return f_win(fn.sum(fn.cast(~col, sa_types.Integer()))) == 0

    return f

sql_median = lambda _, col: fn.percentile_cont(col, .5)

scalar = extend_base(
    BigqueryColumn,
    __floordiv__  = sql_floordiv,
    __rfloordiv__ = annotate(lambda _, x, y: sql_floordiv(y, x), result_type="float"),

    __mod__       = lambda _, x, y: sql.func.mod(x, y),
    mod           = lambda _, x, y: sql.func.mod(x, y),
    __rmod__      = lambda _, x, y: sql.func.mod(y, x),
    rmod          = lambda _, x, y: sql.func.mod(y, x),

    __round__     = annotate(sql_scalar("round"), result_type = "float"),
    round         = annotate(sql_scalar("round"), result_type = "float"),

    # date ----
    **{
      # bigquery has Sunday as 1, pandas wants Monday as 0
      "dt.dayofweek":        lambda _, col: fn.extract("DAYOFWEEK", col) - 2,
      "dt.dayofyear":        sql_extract("DAYOFYEAR"),
      "dt.month_name":       lambda _, col: fn.format_date("%B", col),
      "dt.week":             sql_extract("ISOWEEK"),
      "dt.weekday":          lambda _, col: fn.extract("DAYOFWEEK", col) - 2,
      "dt.weekofyear":       sql_extract("ISOWEEK")
    },
    **{
      "str.contains": sql_str_contains,
      "str.replace": sql_str_replace,
    }
    )

aggregate = extend_base(
    BigqueryColumnAgg,
    # NOTE: bigquery has an all() func, but it's not an aggregate
    any      = sql_any(),
    all      = sql_all(),
    count    = lambda _, col: fn.count(col),
    median   = sql_not_impl(QUANTILE_ERROR),
    nunique  = lambda _, col: fn.count(fn.distinct(col)),
    quantile = sql_not_impl(QUANTILE_ERROR),
    size     = lambda _, col: fn.count("*"),
    std       = sql_agg("stddev"),
    sum      = sql_agg("sum"),
    var      = sql_agg("variance"),
    )

window = extend_base(
    BigqueryColumn,
    any      = sql_any(window = True),
    all      = sql_all(window = True),
    count    = lambda _, col: AggOver(fn.count(col)),
    cumsum   = win_cumul("sum"),
    median   = lambda _, col: RankOver(fn.percentile_cont(col, .5)),
    nunique  = lambda _, col: AggOver(fn.count(fn.distinct(col))),
    quantile = lambda _, col, q: RankOver(fn.percentile_cont(col, q)),
    std       = win_agg("stddev"),
    sum      = win_agg("sum"),
    size     = lambda _, col: AggOver(fn.count("*")),
    var      = win_agg("variance"),
    )


translator = SqlTranslator.from_mappings(
        BigqueryColumn, BigqueryColumnAgg
        )



from pandas.tseries.offsets import Nano, Day, Hour, MonthEnd
from pandas import Timestamp
import pandas as pd
from pandas.core.arrays import DatetimeArray

import pandas.testing as tm
from pandas import offsets
from siuba.experimental.datetime import floor_date, ceil_date

import pytest

# TODO: currently testing only DatetimeIndex, also need to test PeriodIndex
#       and the array classes

def check_floor_ceil(src, offset, unit):
    # Test DatetimeIndex
    res_low = floor_date(src + offset, unit)
    res_hi  = ceil_date(src - offset, unit)

    tm.assert_index_equal(res_low, src)
    tm.assert_index_equal(res_hi, src)

    # Test series ----
    ser_src = pd.Series(src)
    ser_low = floor_date(ser_src + offset, unit)
    ser_hi =  ceil_date(ser_src - offset, unit)

    tm.assert_series_equal(ser_low, ser_src)
    tm.assert_series_equal(ser_hi, ser_src)

@pytest.mark.parametrize("unit", [
    "S",
    "D"
    ])
def test_tick(unit):
    src = pd.DatetimeIndex(["2020-01-02"])
    check_floor_ceil(src, Nano(), unit)


@pytest.mark.parametrize("offset", [Nano(), Day(), Hour()])
def test_nonperiodic_month(offset):
    src  = pd.DatetimeIndex(["2020-02-01"]) 
    check_floor_ceil(src, offset, "MS")


@pytest.mark.parametrize("offset", [Nano(), Day(), Hour()])
def test_floor_nonperiodic_month_end(offset):
    src  = pd.DatetimeIndex(["2020-01-31"]) 
    check_floor_ceil(src, offset, "M")


@pytest.mark.parametrize("offset", [Nano(), Day(), Hour(), MonthEnd()])
def test_floor_nonperiodic_year(offset):
    src = pd.DatetimeIndex(["2020-01-01"]) 
    check_floor_ceil(src, offset, "AS")


@pytest.mark.parametrize("offset", [Nano(), Day(), Hour(), MonthEnd()])
def test_floor_nonperiodic_year_end(offset):
    src = pd.DatetimeIndex(["2019-12-31"]) 
    check_floor_ceil(src, offset, "A")

def test_floor_anchor():
    src = pd.DatetimeIndex(["2020-02-01"])
    res = floor_date(src, "MS")

    tm.assert_index_equal(res, src)


def test_floor_idempotent():
    src = pd.DatetimeIndex(["2020-02-01"])
    res = floor_date(floor_date(src + Day(), "MS"), "MS")

    tm.assert_index_equal(res, src)


def test_ceil_idempotent():
    src = pd.DatetimeIndex(["2020-02-01"])
    res = ceil_date(ceil_date(src - Day(), "MS"), "MS")

    tm.assert_index_equal(res, src)

# Period ----

@pytest.mark.parametrize("unit", [
    "S",
    "D",
    "M"
    ])
def test_period_floor_tick(unit):
    src = pd.DatetimeIndex(["2020-01-02"])
    res = floor_date((src + Nano()).to_period("ns"), unit)
    tm.assert_index_equal(res, src.to_period(unit))


def test_period_ceil_fail():
    with pytest.raises(NotImplementedError):
        ceil_date(pd.PeriodIndex(["2020-01-01"], freq = "D"), "D")



"""
Base dialect for translating column operations to SQL.

A dialect requires three pieces:

    1. Classes representing column data under normal and aggregate settings.
       (e.g. SqlColumn, SqlColumnAgg).
    2. Functions for creating the sqlalchemy clause corresponding to an 
       operation.
       (e.g. a function for taking the standard dev of a column).
    3. A module-level variable named translator. This should be a class with
       a translate method, that takes a column expression and returns a 
       sqlalchemy clause. Ideally using translate.SqlTranslator.
       (e.g. SqlTranslator.from_mapping(...))

"""

# NOTE: this file should be an example that could live outside siuba, so
# we (1) use full import paths, (2) define everything a new backend would need
# here.
from functools import partial

from sqlalchemy import sql
from sqlalchemy import types as sa_types
from sqlalchemy.sql import func as fn
from sqlalchemy.sql.elements import ColumnClause

from siuba.sql.translate import (
        extend_base,
        win_absent,
        win_over,
        win_cumul,
        win_agg,
        sql_agg,
        sql_scalar,
        sql_colmeth,
        sql_ordered_set,
        sql_not_impl,
        annotate,
        RankOver,
        CumlOver,
        SqlTranslator,
        FunctionLookupBound
        )

from siuba.sql.utils import SQLA_VERSION


# =============================================================================
# Column data classes 
# =============================================================================

from siuba.sql.translate import SqlColumn, SqlColumnAgg

# =============================================================================
# Custom translations
# =============================================================================

# Computation -----------------------------------------------------------------

def sql_func_diff(_, col, periods = 1):
    if periods > 0:
        return CumlOver(col - sql.func.lag(col, periods))
    elif periods < 0:
        return CumlOver(col - sql.func.lead(col, abs(periods)))

    raise ValueError("periods argument to sql diff cannot be 0")

def sql_func_floordiv(_, x, y):
    if SQLA_VERSION >= (2, 0, 0):
        return sql.cast(x // y, sa_types.Integer())

    return sql.cast(x / y, sa_types.Integer())

def sql_func_rank(_, col):
    # see https://stackoverflow.com/a/36823637/1144523
    min_rank = RankOver(sql.func.rank(), order_by = col)
    to_mean = (RankOver(sql.func.count(), partition_by = col) - 1) / 2.0

    return min_rank + to_mean


# Datetime -------------------------------------------------------------------

from . import _dt_generics as _dt

def sql_extract(name):
    """Return function that extracts named component from column.

    E.g. to produce EXTRACT(MONTH, some_column)
    """
    return lambda _, col: fn.extract(name, col)


def sql_func_days_in_month(codata, col):
    return fn.extract('day', _dt.sql_func_last_day_in_period(codata, col, 'month'))


def sql_func_extract_dow_monday(_, col):
    # make monday = 0 rather than sunday
    monday0 = sql.cast(sql.func.extract('dow', col) + 6, sa_types.Integer) % 7
    # cast to numeric, since that's what extract('dow') returns
    return sql.cast(monday0, sa_types.Numeric)


def sql_func_floor_date(_, col, unit):
    # see https://www.postgresql.org/docs/9.1/functions-datetime.html#FUNCTIONS-DATETIME-TRUNC
    # valid values: 
    #   microseconds, milliseconds, second, minute, hour,
    #   day, week, month, quarter, year, decade, century, millennium
    # TODO: implement in siuba.dply.lubridate
    return fn.date_trunc(unit, col)


# Strings ---------------------------------------------------------------------

def sql_str_strip(name):
    
    strip_func = getattr(fn, name)
    def f(_, col, to_strip = " \t\n\v\f\r"):
        return strip_func(col, to_strip)

    return f

def sql_func_capitalize(_, col):
    first_char = fn.upper(fn.left(col, 1)) 
    rest = fn.right(col, fn.length(col) - 1)
    return sql.functions.concat(first_char, rest)

def sql_str_cat(_, col, others=None, sep=None, na_rep=None, join=None):
    if sep is not None:
        raise NotImplementedError("sep argument not supported for sql cat")

    if na_rep is not None:
        raise NotImplementedError("na_rep argument not supported for sql cat")

    if join is not None:
        raise NotImplementedError("join argument not supported for sql cat")

    if isinstance(others, (list, tuple)):
        raise NotImplementedError("others argument must be a single column for sql cat")

    return sql.functions.concat(col, others)

# Numpy ufuncs ----------------------------------------------------------------
# symbolic objects have a generic dispatch for when _.__array_ufunc__ is called,
# in order to support things like np.sqrt(_.x). In theory this wouldn't be crazy
# to support, but most ufuncs have existing pandas methods already.

from siuba.siu.symbolic import array_ufunc, array_function

_f_err = FunctionLookupBound("Numpy function sql translation (e.g. np.sqrt) not supported.")

array_ufunc.register(SqlColumn, _f_err)
array_function.register(SqlColumn, _f_err)


# Misc implementations --------------------------------------------------------

def sql_func_astype(_, col, _type):
    mappings = {
            str: sa_types.Text,
            'str': sa_types.Text,
            int: sa_types.Integer,
            'int': sa_types.Integer,
            float: sa_types.Float,
            'float': sa_types.Float,
            bool: sa_types.Boolean,
            'bool': sa_types.Boolean
            }
    try:
        sa_type = mappings[_type]
    except KeyError:
        raise ValueError("sql astype currently only supports type objects: str, int, float, bool")
    return sql.cast(col, sa_type)


# Annotations -----------------------------------------------------------------

def req_bool(f):
    return annotate(f, input_type = "bool")


# =============================================================================
# Base translation mappings
# =============================================================================

extend_base(SqlColumn,
    # infix ----
    __add__       = sql_colmeth("__add__"),
    __and__       = req_bool(sql_colmeth("__and__")),
    __eq__        = sql_colmeth("__eq__"),
    __floordiv__  = sql_func_floordiv,
    __ge__        = sql_colmeth("__ge__"),
    __gt__        = sql_colmeth("__gt__"),
    __invert__    = req_bool(sql_colmeth("__invert__")),
    __le__        = sql_colmeth("__le__"),
    __lt__        = sql_colmeth("__lt__"),
    __mod__       = sql_colmeth("__mod__"),
    __mul__       = sql_colmeth("__mul__"),
    __ne__        = sql_colmeth("__ne__"),
    __neg__       = sql_colmeth("__neg__"),
    __or__        = req_bool(sql_colmeth("__or__")),
    __pos__       = sql_not_impl(),
    __pow__       = sql_not_impl(),
    __radd__      = sql_colmeth("__radd__"),
    __rand__      = req_bool(sql_colmeth("__rand__")),
    __rfloordiv__ = lambda _, x, y: sql_func_floordiv(y, x),
    __rmod__      = sql_colmeth("__rmod__"),
    __rmul__      = sql_colmeth("__rmul__"),
    __ror__       = req_bool(sql_colmeth("__ror__")),
    __round__     = sql_scalar("round"),
    __rpow__      = sql_not_impl(),
    __rsub__      = sql_colmeth("__rsub__"),
    __rtruediv__  = sql_colmeth("__rtruediv__"),
    #__rxor__      = sql_colmeth("__rxor__"),
    __sub__       = sql_colmeth("__sub__"),
    __truediv__   = sql_colmeth("__truediv__"),
    #__xor__       = sql_colmeth("__xor__"), 


    # infix methods ----

    add           = sql_colmeth("__add__"),        
    #and          =
    div           = sql_colmeth("__truediv__"),
    divide        = sql_colmeth("__truediv__"),
    #divmod       = 
    eq            = sql_colmeth("__eq__"),         
    #floordiv     = sql_colmeth("__floordiv__"),         
    ge            = sql_colmeth("__ge__"),     
    gt            = sql_colmeth("__gt__"),         
    le            = sql_colmeth("__le__"),         
    lt            = sql_colmeth("__lt__"),        
    mod           = sql_colmeth("__mod__"),        
    mul           = sql_colmeth("__mul__"),         
    multiply      = sql_colmeth("__mul__"),        
    ne            = sql_colmeth("__ne__"),         
    pow           = sql_not_impl(),
    radd          = sql_colmeth("__radd__"),       
    rdiv          = sql_colmeth("__rtruediv__"),
    #rdivmod      = 
    #rfloordiv    = sql_colmeth("__pow__"),        
    rmod          = sql_colmeth("__rmod__"),       
    rmul          = sql_colmeth("__rmul__"),       
    round         = sql_scalar("round"),
    rpow          = sql_not_impl(),     
    rsub          = sql_colmeth("__rsub__"),       
    rtruediv      = sql_not_impl(),
    sub           = sql_colmeth("__sub__"),   
    subtract      = sql_colmeth("__sub__"),      
    truediv       = sql_colmeth("__truediv__"),        


    # computation ---- 

    abs                     = sql_scalar("abs"),
    between                 = sql_colmeth("between"),
    clip                    = lambda _, col, low, upp: fn.least(fn.greatest(col, low), upp),
    isin                    = sql_colmeth("in_"),


    # strings ----


    **{
      # TODO: check generality of trim functions, since MYSQL overrides
      "str.capitalize"    : sql_func_capitalize,
      "str.cat"           : sql_str_cat,
      #"str.center"        :,
      #"str.contains"      :,
      #"str.count"         :,
      #"str.encode"        :,
      "str.endswith"      : sql_colmeth("endswith"),
      #"str.find"          :,
      #"str.findall"       :,
      #"str.isalnum"       :,
      #"str.isalpha"       :,
      #"str.isdecimal"     :,
      #"str.isdigit"       :,
      "str.islower"       : lambda _, col: col == sql.func.lower(col),
      #"str.isnumeric"     :,
      #"str.isspace"       :,
      #"str.istitle"       :,
      #"str.isupper"       :,
      "str.len"           : lambda _, col: sql.func.length(col),
      #"str.ljust"         :,
      "str.lower"         : lambda _, col: sql.func.lower(col),
      "str.lstrip"        : sql_str_strip("ltrim"),
      #"str.match"         :,
      #"str.pad"           :,
      #"str.replace"       :,
      #"str.rfind"         :,
      #"str.rjust"         :,
      #"str.rsplit"        :,
      "str.rstrip"        : sql_str_strip("rtrim"),
      #"str.slice"         :,
      #"str.slice_replace" :,
      #"str.split"         :,
      "str.startswith"    : sql_colmeth("startswith"),
      "str.strip"         : sql_str_strip("trim"),
      #"str.swapcase"      :,
      "str.title"         : lambda _, col: sql.func.initcap(col),
      "str.upper"         : lambda _, col: sql.func.upper(col),
      #"str.wrap"          :,
    },


    # datetime ----


    **{
      #"dt.ceil"             :
      #"dt.date"             :
      "dt.day"              : sql_extract("day"),
      #"dt.day_name"         :
      "dt.dayofweek"        : sql_func_extract_dow_monday,
      "dt.dayofyear"        : sql_extract("doy"),
      #"dt.days"             :
      "dt.days_in_month"    : sql_func_days_in_month,
      "dt.daysinmonth"      : sql_func_days_in_month,
      #"dt.floor"            :
      "dt.hour"             : sql_extract("hour"),
      #"dt.is_leap_year"     :
      "dt.is_month_end"     : partial(_dt.sql_is_last_day_of, period="month"),
      "dt.is_month_start"   : partial(_dt.sql_is_first_day_of, period="month"),
      "dt.is_quarter_end"   : partial(_dt.sql_is_last_day_of, period="quarter"),
      "dt.is_quarter_start" : partial(_dt.sql_is_first_day_of, period="quarter"),
      "dt.is_year_end"      : partial(_dt.sql_is_last_day_of, period="year"),
      "dt.is_year_start"    : partial(_dt.sql_is_first_day_of, period="year"),
      #"dt.microsecond"      :
      #"dt.microseconds"     :
      "dt.minute"           : sql_extract("minute"),
      "dt.month"            : sql_extract("month"),
      #"dt.month_name"       :
      #"dt.nanosecond"       :
      #"dt.nanoseconds"      :
      #"dt.normalize"        :
      "dt.quarter"          : sql_extract("quarter"),
      #"dt.qyear"            :
      #dt.round            =
      "dt.second"           : sql_extract("second"),
      #"dt.seconds"          :
      #"dt.strftime"         :
      #"dt.time"             :
      #"dt.timetz"           :
      #"dt.to_period"        :
      #"dt.to_pydatetime"    :
      #"dt.to_pytimedelta"   :
      #"dt.to_timestamp"     :
      #"dt.total_seconds"    :
      #dt.tz_convert       =
      #dt.tz_localize      =
      "dt.week"             : sql_extract("week"),
      "dt.weekday"          : sql_func_extract_dow_monday,
      "dt.weekofyear"       : sql_extract("week"),
      "dt.year"             : sql_extract("year"),
      #"dt.freq" : 
      #"dt.tz"   :
      },


    # datetime methods not on accessor ----

    #asfreq            = 
    #between_time      = TODO


    # Missing values ----

    fillna      = lambda _, x, y: sql.functions.coalesce(x,y),
    isna        = sql_colmeth("is_", None),
    isnull      = sql_colmeth("is_", None),
    notna       = lambda _, col: ~col.is_(None),
    notnull     = lambda _, col: ~col.is_(None),

    # Misc ---
    #replace       =  # TODO
    astype = sql_func_astype,
    #where
)


extend_base(SqlColumn,

    # computation ----
    #autocorr                = 
    cummax                  = win_cumul("max"),
    cummin                  = win_cumul("min"),
    #cumprod                 = 
    cumsum                  = annotate(win_cumul("sum"), result_type = "variable"),
    diff                    = sql_func_diff,
    #is_monotonic            = 
    #is_monotonic_decreasing = 
    #is_monotonic_increasing = 
    #pct_change              = TODO(?)
    rank                    = sql_func_rank,

    # computation (strict aggregates)
    #all = #TODO(pg): all = sql_aggregate("BOOL_AND", "all")
    #any = #TODO(pg): any = sql_aggregate("BOOL_OR", "any"),
    #corr = # TODO(pg)
    count = win_agg("count"),
    #cov = 
    #is_unique = # TODO(low)
    #kurt = 
    #kurtosis = 
    #mad = 
    max = win_agg("max"),
    mean = win_agg("avg"),
    #median = 
    min = win_agg("min"),
    nunique = win_absent("nunique"),
    #prod = 
    #product = 
    quantile = sql_ordered_set("percentile_cont", is_analytic=True),
    #sem = 
    #skew = 
    #std =  # TODO(pg)
    sum = annotate(win_agg("sum"), result_type = "variable"),
    #var = # TODO(pg)



    # datetime not on accessor ----
    #asof              =
    #shift             =
    #tshift            =


    # missing values ----

    #bfill       = TODO: pg
    #ffill       = TODO: pg

    # index (strict aggregates) ----

    #equals =  #TODO(pg) combine == and all
    #idxmax =  #TODO?
    #idxmin = 


    # attributes (strict aggregates) ----

    #empty = 
    #hasnans = # TODO
    #memory_usage = 
    #nbytes = 
    size = win_absent("count"),
)


# =============================================================================
# Aggregate functions
# =============================================================================

extend_base(SqlColumnAgg,
    # infix methods ----

    #dot = sql_not_impl(),

    # computation ----

    #all = #TODO(pg): all = sql_aggregate("BOOL_AND", "all")
    #any = #TODO(pg): any = sql_aggregate("BOOL_OR", "any"),
    #corr = # TODO(pg)
    count = lambda _, col: sql.func.count(),
    #cov = 
    #is_unique = # TODO(low)
    #kurt = 
    #kurtosis = 
    #mad = 
    max = sql_agg("max"),
    mean = sql_agg("avg"),
    #median = 
    min = sql_agg("min"),
    nunique = lambda _, col: sql.func.count(sql.func.distinct(col)),
    #prod = 
    #product = 
    quantile = sql_ordered_set("percentile_cont"),
    #sem = 
    #skew = 
    #std =  # TODO(pg)
    sum = annotate(sql_agg("sum"), result_type = "float"),
    #var = # TODO(pg)

    # index ----

    #equals =  #TODO(pg) combine == and all
    #idxmax =  #TODO?
    #idxmin = 


    # attributes ----

    #empty = 
    #hasnans = # TODO
    #memory_usage = 
    #nbytes = 
    size = sql_agg("count"),
)


# =============================================================================
# Base with no windows implemented (for e.g. sqlite)
# =============================================================================

# based on https://github.com/tidyverse/dbplyr/blob/master/R/backend-.R
base_nowin = dict(
    cummax                  = win_absent("max"),
    cummin                  = win_absent("min"),
    cumsum                  = win_absent("sum"),
    diff                    = win_absent("diff"),
    rank                    = win_absent("rank"),
    count                   = win_absent("count"),
    max =                     win_absent("max"),
    mean = win_absent("avg"),
    min = win_absent("min"),
    nunique = win_absent("nunique"),
    quantile =  win_absent("quantile"),
    sum = win_absent("sum"),
    size = win_absent("count"),
)


translator = SqlTranslator.from_mappings(
        SqlColumn, SqlColumnAgg
        )



import json
import yaml
import pkg_resources
import pandas as pd

from siuba.ops import ALL_OPS

from siuba.siu import FunctionLookupBound
from siuba.sql.utils import get_dialect_translator

SQL_BACKENDS = ["postgresql", "redshift", "sqlite", "mysql", "bigquery", "snowflake", "duckdb"]
ALL_BACKENDS = SQL_BACKENDS + ["pandas"]

methods = pd.DataFrame(
        [{"full_name": k, **vars(f.operation)} for k, f in ALL_OPS.items()]
        )


# SQL spec construction -------------------------------------------------------

def read_dialect(name):
    translator = get_dialect_translator(name)

    support = []
    for k in ALL_OPS:
        support.append(read_sql_op(k, name, translator))

    df_support = pd.DataFrame(support)

    return df_support



def read_sql_op(name, backend, translator):
    # TODO: MC-NOTE - cleanup this code
    from siuba.siu.visitors import CodataVisitor, FunctionLookupError
    from siuba.ops.utils import Operation
    co_win = CodataVisitor(translator.window.dispatch_cls)
    co_agg = CodataVisitor(translator.aggregate.dispatch_cls)

    disp_win = translator.window.local[name]
    disp_agg = translator.aggregate.local[name]

    try:
        f_win = co_win.validate_dispatcher(disp_win, strict=False)
        if isinstance(f_win, FunctionLookupBound):
            win_supported = False
        elif disp_win.dispatch(object) is f_win:
            win_supported = False
        else:
            win_supported = True
    except FunctionLookupError:
        f_win = None
        win_supported = False


    try:
        f_agg = co_agg.validate_dispatcher(disp_agg)
        if isinstance(f_agg, FunctionLookupBound):
            agg_supported = False
        else:
            agg_supported = True
    except FunctionLookupError:
        agg_supported = False

    # window functions should be a superset of agg functions
    if f_win is None and agg_supported:
        raise Exception("agg functions in %s without window funcs: %s" %(backend, name))

    if win_supported and not agg_supported:
        flags = "no_aggregate"
    elif agg_supported and not win_supported:
        flags = "no_mutate"
    else:
        flags = ""

    if win_supported or agg_supported:
        metadata = getattr(f_win, "operation", {})
        if isinstance(metadata, Operation):
            metadata = {**vars(metadata)}
        meta = {"is_supported": True, "flags": flags, **metadata}

    else:
        meta = {"is_supported": False, "flags": flags}
    
    return {"full_name": name, "backend": backend, "metadata": meta}


# pandas spec construction ----------------------------------------------------

def read_pandas_ops():
    from siuba.experimental.pd_groups.groupby import SeriesGroupBy

    all_meta = []
    for k, v in ALL_OPS.items():
        has_impl = not isinstance(v.dispatch(SeriesGroupBy), FunctionLookupBound)
        all_meta.append({
            "full_name": k,
            "metadata": dict(is_supported = has_impl),
            "backend": "pandas"}
            )

    return pd.DataFrame(all_meta)

# process examples ------------------------------------------------------------
# examples currently have form like _ + _, or _.corr(_)
# these need to be converted to _.x + _.y, and _x.corr(_.y)

def enrich_spec_entry(entry):
    from siuba.siu import _, strip_symbolic
    accessors = ['str', 'dt', 'cat', 'sparse']
    expr = strip_symbolic(eval(entry["example"], {"_": _}))

    accessor = [ameth for ameth in accessors if ameth in expr.op_vars()] + [None]

    tmp = {
            **entry,
            'is_property': expr.func == "__getattr__",
            'expr_frame': replace_meta_args(expr, _.x, _.y, _.z),
            'expr_series': expr,
            'accessor': accessor[0],
            }

    return tmp

def replace_meta_args(call, *args):
    from siuba.siu import strip_symbolic, MetaArg
    replacements = map(strip_symbolic, args)
    
    def replace_next(node):
        # swap in first replacement for meta arg (when possible)
        if isinstance(node, MetaArg):
            return next(replacements)
        
        # otherwise, recurse into node
        args, kwargs = node.map_subcalls(replace_next)

        # return a copy of node
        return node.__class__(node.func, *args, **kwargs)

    return replace_next(call)


# Main ========================================================================

sql_methods    = pd.concat(list(map(read_dialect, SQL_BACKENDS)))
pandas_methods = pd.DataFrame(read_pandas_ops())

wide_backends = (
        pd.concat([sql_methods, pandas_methods])
        .pivot(index="full_name", columns="backend", values="metadata")
        )

full_methods = methods.merge(wide_backends, how = "left", on = "full_name")

# replace NA entries--but pandas' fillna has custom behavior around dicts, so
# need to use this terrible hack
def set_default_support(d):
    if d.get("is_supported"):
        return {**d, "support": "supported"}

    return {**d, "is_supported": False, "support": "maydo"}

for be_name in ALL_BACKENDS:
    full_methods[be_name] = full_methods[be_name].apply(set_default_support)


# Nest backends from each being a column, to a single column of dicts
df_spec = (full_methods
        .assign(backends = lambda _: _[ALL_BACKENDS].to_dict(orient = "records"))
        .drop(columns = ALL_BACKENDS)
        )

fname_spec = pkg_resources.resource_filename("siuba.ops.support", "examples.yml")
with open(fname_spec, "r") as f:
    orig_spec = yaml.safe_load(f)

df_spec["example"] = df_spec.full_name.apply(lambda x: orig_spec[x])

# roundtrip through JSON to get rid of custom numpy types
# probably a much better way to do this
json_spec = df_spec.set_index("full_name").to_json(orient = "index")
raw_spec = json.loads(json_spec)

spec = {k: enrich_spec_entry(entry) for k, entry in raw_spec.items()} 





import itertools
import operator

from abc import ABC
from collections.abc import Mapping

# TODO: symbolic formatting: __add__ -> "+"

BINARY_LEVELS = {
        "__add__": 2,
        "__sub__": 2,

        "__mul__": 1,
        "__matmul__": 1,
        "__truediv__": 1,
        "__floordiv__": 1,
        "__mod__": 1,
        "__divmod__": 1,

        "__pow__": 0,

        "__lshift__": 3,
        "__rshift__": 3,

        "__and__": 4,
        "__xor__": 4,
        "__or__": 4,
        "__gt__": 5,
        "__lt__": 5,
        "__eq__": 5,
        "__ne__": 5,
        "__ge__": 5,
        "__le__": "5",

        "__getattr__": 0,
        "__getitem__": 0,
        }

BINARY_OPS = {
        "__add__": "+",
        "__sub__": "-",
        "__mul__": "*",
        "__matmul__": "@",
        "__truediv__": "/",
        "__floordiv__": "//",
        "__mod__": "%",
        "__divmod__": "IDK",
        "__pow__": "**",
        "__lshift__": "<<",
        "__rshift__": ">>",
        "__and__": "&",
        "__xor__": "^",
        "__or__": "|",
        "__gt__": ">",
        "__lt__": "<",
        "__eq__": "==",
        "__ne__": "!=",
        "__ge__": ">=",
        "__le__": "<=",
        "__getattr__": ".",
        "__getitem__": "[",
        }

UNARY_OPS = {
        "__invert__": "~",
        "__neg__": "-",
        "__pos__": "+",
        "__abs__": "ABS => "
        }

# TODO: can it just be put in binary ops? Special handling in Symbolic class?
ALL_OPS = {**BINARY_OPS, **UNARY_OPS}

BINARY_RIGHT_OPS = {}
for k, v in BINARY_OPS.items():
    BINARY_RIGHT_OPS[k.replace("__", "__r", 1)] = k

for k, v in BINARY_LEVELS.copy().items():
    BINARY_LEVELS[k.replace("__", "__r", 1)] = v

# Functions
# ==============================================================================

def str_to_getitem_call(x):
    return Call("__getitem__", MetaArg("_"), x)
    

# Calls
# =============================================================================

class Call:
    """Represent python operations.

    This class is responsible for representing the pieces of a python expression,
    as a function, along with its args and kwargs.

    For example, "some_object.a" would be represented at the function "__getattr__",
    with the args `some_object`, and `"a"`.

    Parameters
    ----------
    func :
        Name of the function called. Class methods are represented using the names
        they have when defined on the class.
    *args :
        Arguments the function call uses.
    **kwargs :
        Keyword arguments the function call uses.

    Examples
    --------
    >>> Call("__add__", 1, 1)
    (1 + 1)

    See Also
    --------
    siuba.siu.Symbolic : Helper class for creating Calls.


    """


    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        """Return a (best guess) python code representation of the Call.

        Note that this is not necessarily valid python code (e.g. if a python
        function is passed as a call argument).
        """
        # TODO: format binary, unary, call, associative
        if self.func in BINARY_OPS:
            op_repr = BINARY_OPS[self.func]
            fmt = "({args[0]} {func} {args[1]})"
        elif self.func in UNARY_OPS:
            op_repr = UNARY_OPS[self.func]
            fmt = "({func}{args[0]})"
        elif self.func == "getattr":
            op_repr = "."
            fmt = "({args[0]}{func}{args[1]})"
        else:
            op_repr, *arg_str = map(repr, self.args)
            kwarg_str = (str(k) + " = " + repr(v) for k,v in self.kwargs.items())

            combined_arg_str = ",".join(itertools.chain(arg_str, kwarg_str))
            fmt = "{}({})".format(op_repr, combined_arg_str)
            return fmt

        return fmt.format(
                    func = op_repr or self.func,
                    args = self.args,
                    kwargs = self.kwargs
                    )

    def __call__(self, x):
        """Evaluate a call over some context and return the result.

        Note that subclasses like MetaArg, simply return the context, so that a call
        acts like a unary function, with ``x`` as its argument.

        Parameters
        ----------
        x :
            Object passed down the call tree as context.

        Examples
        --------
        >>> expr = Call("__add__", MetaArg("_"), 2)
        >>> expr(1)   # 1 + 2
        3

        
        """
        args, kwargs = self.map_subcalls(self.evaluate_calls, args = (x,))
        inst, *rest = args

        #inst, *rest = (self.evaluate_calls(arg, x) for arg in self.args)
        #kwargs = {k: self.evaluate_calls(v, x) for k, v in self.kwargs.items()}
        
        # TODO: temporary workaround, for when only __get_attribute__ is defined
        if self.func == "__getattr__":
            return getattr(inst, *rest)
        elif self.func == "__getitem__":
            return operator.getitem(inst, *rest)
        elif self.func == "__call__":
            return getattr(inst, self.func)(*rest, **kwargs)

        # in normal case, get method to call, and then call it
        f_op = getattr(operator, self.func)
        return f_op(inst, *rest, **kwargs)

    # TODO: type checks will be very useful here. Will need to import symbolic.
    # Let's do this once types are in a _typing.py submodule.
    def __rshift__(self, x):
        """Create a"""
        from .symbolic import strip_symbolic

        stripped = strip_symbolic(x)

        if isinstance(stripped, Call):
            return self._construct_pipe(self, x)

        raise TypeError()

    def __rrshift__(self, x):
        from .symbolic import strip_symbolic
        if isinstance(strip_symbolic(x), (Call)):
            # only allow non-calls (i.e. data) on the left.
            raise TypeError()
 
        return self(x)

    @staticmethod
    def evaluate_calls(arg, x):
        if isinstance(arg, Call): return arg(x)

        return arg

    def copy(self) -> "Call":
        """Return a copy of this call object.

        Note that copies are made of child calls, but not their arguments.
        """
        args, kwargs = self.map_subcalls(lambda child: child.copy())
        return self.__class__(self.func, *args, **kwargs)

    def map_subcalls(self, f, args = tuple(), kwargs = None):
        """Call a function on all child calls.

        Parameters
        ----------
        f :
            A function to call on any child calls.
        args:
            Optional position arguments to pass to ``f``.
        kwargs:
            Optional keyword arguments to pass to ``f``.

        Returns
        -------
        A tuple of (new_args, new_kwargs) that can be used to recreate the original
        Call object with transformed (copies of) child nodes.

        See Also
        --------
        copy : Recursively calls map_subcalls to clone a call tree.

        """
        if kwargs is None: kwargs = {}

        new_args = tuple(f(arg, *args, **kwargs) if isinstance(arg, Call) else arg for arg in self.args)
        new_kwargs = {k: f(v, *args, **kwargs) if isinstance(v, Call) else v for k,v in self.kwargs.items()}

        return new_args, new_kwargs

    def map_replace(self, f):
        args, kwargs = self.map_subcalls(f)
        return self.__class__(self.func, *args, **kwargs)

    def iter_subcalls(self, f):
        yield from iter(arg for arg in self.args if instance(arg, Call))
        yield from iter(v for k,v in self.kwargs.items() if isinstance(v, Call))

    def op_vars(self, attr_calls = True):
        """Return set of all variable names used in Call
        
        Args:
            attr_calls: whether to include called attributes (e.g. 'a' from _.a())
        """
        varnames = set()

        op_var = self._get_op_var()
        if op_var is not None:
            varnames.add(op_var)

        if (not attr_calls
            and self.func == "__call__"
            and isinstance(self.args[0], Call)
            and self.args[0].func == "__getattr__"
            ):
            # skip obj, since it fetches an attribute this node is calling
            prev_obj, prev_attr = self.args[0].args
            all_args = itertools.chain([prev_obj], self.args[1:], self.kwargs.values())
        else:
            all_args = itertools.chain(self.args, self.kwargs.values())


        for arg in all_args:
            if isinstance(arg, Call):
                varnames.update(arg.op_vars(attr_calls = attr_calls))

        return varnames
    
    def _get_op_var(self):
        if self.func in ("__getattr__", "__getitem__") and isinstance(self.args[1], str):
            return self.args[1]

    def obj_name(self):
        obj = self.args[0]
        if isinstance(obj, Call):
            if obj.func == "__getattr__":
                return obj.args[0]
        elif hasattr(obj, '__name__'):
            return obj.__name__

        return None

    @staticmethod
    def _construct_pipe(lhs, rhs):
        if isinstance(lhs, PipeCall):
            lh_args = lhs.args

            # ensure we don't keep adding MetaArg to the left when
            # combining two pipes
            if lh_args and isinstance(lh_args[0], MetaArg):
                lh_args = lh_args[1:]
        else:
            lh_args = [lhs]

        if isinstance(rhs, PipeCall):
            rh_args = rhs.args

            # similar to above, but for rh args
            if rh_args and isinstance(rh_args[0], MetaArg):
                rh_args = rh_args[1:]
        else:
            rh_args = [rhs]

        return PipeCall(MetaArg("_"), *lh_args, *rh_args)


class Lazy(Call):
    """Lazily return calls rather than evaluating them."""
    def __init__(self, func, arg = None):
        if arg is None:
            self.func = "<lazy>"
            self.args = [func]
        else:
            # will happen in generic node calls, e.g. self.__class__(self.func, ...)
            self.func = func
            self.args = [arg]

        self.kwargs = {}

    def __call__(self, x, *args, **kwargs):
        return self.args[0]


class _Isolate(Lazy):
    """Lazily return calls, and do dispatch visitors."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func = "<isolate>"

    def map_subcalls(self, f, args = tuple(), kwargs = None):
        return self.args, {**self.kwargs}


class UnaryOp(Call):
    """Represent unary call operations."""
    def __repr__(self):
        fmt = "{func}{args[0]}"

        func = UNARY_OPS[self.func]
        return fmt.format(func = func, args = self.args, kwargs = self.kwargs)


class BinaryOp(Call):
    """Represent binary call operations."""

    def __repr__(self):
        return self._repr(reverse = False)


    def _repr(self, reverse = False):
        func_name = self.get_func_name()

        level = BINARY_LEVELS[func_name]
        spaces = "" if level == 0 else " "

        args = self.args
        arg0 = "({args[0]})" if self.needs_paren(args[0]) else "{args[0]}"
        arg1 = "({args[1]})" if self.needs_paren(args[1]) else "{args[1]}"

        # handle binary ops that are not infix operators
        if self.func == "__getitem__":
            suffix = "]"
        else:
            suffix = ""

        # final, formatting
        fmt = arg0 + "{spaces}{func}{spaces}" + arg1 + suffix


        func = BINARY_OPS[func_name]
        if self.func == "__getattr__":
            # use bare string. eg _.a
            fmt_args = [repr(args[0]), args[1]]
        else:
            fmt_args = list(map(repr, args))
        
        if reverse:
            fmt_args = list(reversed(fmt_args))

        return fmt.format(func = func, args = fmt_args, spaces = spaces)

    def get_func_name(self):
        return BINARY_RIGHT_OPS.get(self.func, self.func)
        

    def needs_paren(self, x):
        if isinstance(x, BinaryOp):
            sub_lvl = BINARY_LEVELS[x.get_func_name()]
            level = BINARY_LEVELS[self.get_func_name()]
            if sub_lvl != 0 and sub_lvl > level:
                return True

        return False

class BinaryRightOp(BinaryOp):
    """Represent right associative binary call operations."""

    def __call__(self, x):
        inst, *rest = (self.evaluate_calls(arg, x) for arg in self.args)
        kwargs = {k: self.evaluate_calls(v, x) for k, v in self.kwargs.items()}
        
        # in normal case, get method to call, and then call it
        func_name = BINARY_RIGHT_OPS[self.func]
        f_op = getattr(operator, func_name)

        # TODO: in practice rest only has 1 item, but this is not enforced..
        return f_op(*rest, inst, **kwargs)

    def __repr__(self):
        return self._repr(reverse = True)


class DictCall(Call):
    """Calls representation of dictionary construction.

    Note that this class is important for two reasons:

      * Dictionary literal syntax cannot produce a call.
      * Calls cannot recognize subcalls nested inside containers. E.g. dicts, lists.
    """

    def __init__(self, f, *args, **kwargs):
        # TODO: validation, clean up class
        super().__init__(f, *args, **kwargs)
        self.args = (dict, dict(self.args[1]))

    def map_subcalls(self, f, args = tuple(), kwargs = None):
        if kwargs is None: kwargs = {}

        d = self.args[1]

        # TODO: don't use a giant comprehension
        new_d = {
            f(k, *args, **kwargs) if isinstance(k, Call) else k     # key part
            : f(v, *args, **kwargs) if isinstance(v, Call) else v   # val part
            for k,v in d.items()
            }

        return (self.args[0], new_d), {}

    def __call__(self, x):
        # TODO: note that it will not descend into dict to evaluate calls
        return self.args[1]


# Slice Calls -----------------------------------------------------------------
# note the metaclass SliceOp below, which registers SliceOpExt

class _SliceOpExt(Call):
    """Represent arguments for extended slice syntax.

    E.g. expressions like _[_:, 'a', 1:2, ] use a tuple of many slices.
    """

    def __init__(self, func, *args, **kwargs):
        self.func = "__siu_slice__"

        if kwargs:
            raise ValueError("a slice cannot accept keyword arguments")

        self.args = args
        self.kwargs = {}

    def __repr__(self):
        return ", ".join(map(self._repr_slice, self.args))

    def __call__(self, x):
        args, kwargs = self.map_subcalls(self.evaluate_calls, args = (x,))

        return args

    def map_subcalls(self, f, args=tuple(), kwargs=None):
        if kwargs is None: kwargs = {}

        # evaluate each argument, which can be a slice
        args = tuple(self._apply_slice_entry(f, obj, args, kwargs) for obj in self.args)

        return args, {}

    def op_vars(self, *args, **kwargs):
        def visit_op_vars(node, *args, **kwargs):
            if not isinstance(node, Call):
                return set()

            return node.op_vars(*args, **kwargs)

        results = set()
        for child in self.args:
            # TODO: refactor Call classes to use singledispatch, so we can call
            # op_vars directly on a slice object
            if isinstance(child, slice):
                for piece in (child.start, child.stop, child.step):
                    results.update(visit_op_vars(piece, *args, **kwargs))

            else:
                results.update(visit_op_vars(child, *args, **kwargs))

        return results

    @staticmethod
    def _apply_slice_entry(f, slice_, args, kwargs):
        if not isinstance(slice_, slice):
            # a slice argument can be simple literal. e.g. _['a']
            return f(slice_, *args, **kwargs) if isinstance(slice_, Call) else slice_

        slice_args = (slice_.start, slice_.stop, slice_.step)
        args = [
            f(entry, *args, **kwargs) if isinstance(entry, Call) else entry
            for entry in slice_args
            ]

        return slice(*args)

    @staticmethod
    def _unslice(slice_):
        if isinstance(slice_, slice):
            return (slice_.start, slice_.stop, slice_.step)

        return slice_

    @classmethod
    def _repr_slice(cls, slice_):
        if isinstance(slice_, slice):
            pieces = cls._unslice(slice_)

            # if the last piece is not specified, e.g. :1, then we want to represent
            # it like that, and not the equivalent :1:.
            if pieces[-1] is None:
                pieces = pieces[:-1]
        else:
            pieces = (slice_,)

        return ":".join(repr(x) if x is not None else "" for x in pieces)


class _SliceOpIndex(_SliceOpExt):
    """Special case of slicing, where getitem receives a single slice object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        return res[0]


class SliceOp(ABC):
    """Factory class for representing from single and extended slice calls. 

    Note that it has SliceOpIndex, and SliceOpExt registered as subclasses, so
    that this class can be used rather than the specific implementations.
    """

    def __new__(cls, func, *args, **kwargs):
        # must be constructed in the same way as __getitem__ sees the indexer.
        # that is, a single argument (that can be a tuple)
        if len(args) != 1:
            raise ValueError("SliceOpIndex allows 1 argument, but received %s" % len(args))

        elif isinstance(args[0], tuple):
            # general case, where calling returns a tuple of indexers.
            # e.g. _['a', ], or _[1:, :, :, :]
            return _SliceOpExt(func, *args[0])

        else:
            # special case, where calling returns a single indexer, rather than
            # a tuple of indexers (e.g. _['a'], or _[1:])
            return _SliceOpIndex(func, args[0])


SliceOp.register(_SliceOpExt)


# formulas ---------------------------

class FormulaContext(Mapping):
    def __init__(self, *args, **kwargs):
        self.d = dict(*args, **kwargs)

    def __getitem__(self, k):
        return self.d[k]

    def __iter__(self):
        return iter(self.d)

    def __len__(self):
        return len(self.d)


# Special kinds of call arguments ----
# These functions insure that when using siu expressions generated by _,
# that call.args[0] is always another call. This allows them to trivially
# respond to calling a siu expression, map_subcalls, etc..
#
# In the future, could make a parent class for Call, with a restricted
# set of behavior similar to theirs.
#
# TODO: validate that call.args[0] is a Call in tree visitors?
class MetaArg(Call):
    """Represent an argument, by returning the argument passed to __call__."""

    def __init__(self, func, *args, **kwargs):
        self.func = "_"
        self.args = tuple()
        self.kwargs = {}

    def __repr__(self):
        return self.func

    def __call__(self, x):
        if isinstance(x, FormulaContext):
            return x[self.func]

        return x


# TODO: MetaArg should be a subclass of this?
class FormulaArg(MetaArg):
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = tuple()
        self.kwargs = {}

    def __repr__(self):
        return f"FormulaArg({repr(self.func)})"

    def __call__(self, x: FormulaContext):
        if isinstance(x, FormulaContext):
            return x[self.func]

        raise TypeError(
            f"The formula object {self.func} must receive a FormulaContext when called."
        )



class FuncArg(Call):
    """Represent a function to be called."""

    def __init__(self, func, *args, **kwargs):
        self.func = '__custom_func__'

        if func == '__custom_func__':
            func = args[0]

        self.args = tuple([func])
        self.kwargs = {}

    def __repr__(self):
        return repr(self.args[0])

    def __call__(self, x):
        return self.args[0]


# Pipe ===================================================================================

class PipeCall(Call):
    """
    pipe(df, a, b, c)
    pipe(_, a, b, c)

    should options for first arg be only MetaArg or a non-call?
    """

    def __init__(self, func, *args, **kwargs):
        if isinstance(func, str) and func == "__siu_pipe_call__":
            # it was a mistake to make func the first parameter to Call
            # but basically we need to catch when it is passed, so
            # we can ignore it
            self.func = func
            self.args = args
        else:
            self.func = "__siu_pipe_call__"
            self.args = (func, *args)
        if kwargs:
            raise ValueError("Keyword arguments are not allowed.")
        self.kwargs = {}

    def __call__(self, x=None):
        # Note that most calls map_subcalls to pass in the same data for each argument.
        # In contrast, PipeCall passes data from the prev step to the next.
        crnt_data, *calls = self.args

        if isinstance(crnt_data, MetaArg):
            crnt_data = crnt_data(x)

        for call in calls:
            new_data = call(crnt_data)
            crnt_data = new_data

        return crnt_data
    
    def __repr__(self):
        args_repr = ",".join(map(repr, self.args))
        return f"{type(self).__name__}({args_repr})"



import pandas as pd
import numpy as np
from functools import singledispatch
import itertools

from ..siu import Symbolic, symbolic_dispatch


def _coerce_to_str(x):
    if isinstance(x, (pd.Series, np.ndarray)):
        return x.astype(str)
    elif not np.ndim(x) < 2:
        raise ValueError("np.ndim must be less than 2, but is %s" %np.ndim(x))

    return pd.Series(x, dtype = str)


@symbolic_dispatch(cls = pd.Series)
def str_c(x, *args, sep = "", collapse = None):
    all_args = itertools.chain([x], args)
    strings = list(map(_coerce_to_str, all_args))

    return np.sum(strings, axis = 0)




from siuba.dply.verbs import select, rename, _select_group_renames 
from siuba.dply.tidyselect import VarList, var_select
from siuba.dply.verbs import simple_varname

from pandas import Series

from ..backend import LazyTbl, _warn_missing
from ..utils import lift_inner_cols, _sql_with_only_columns


@select.register(LazyTbl)
def _select(__data, *args, **kwargs):
    # see https://stackoverflow.com/questions/25914329/rearrange-columns-in-sqlalchemy-select-object
    if kwargs:
        raise NotImplementedError(
                "Using kwargs in select not currently supported. "
                "Use _.newname == _.oldname instead"
                )
    last_sel = __data.last_select
    columns = {c.key: c for c in last_sel.inner_columns}

    # same as for DataFrame
    colnames = Series(list(columns))
    vl = VarList()
    evaluated = (arg(vl) if callable(arg) else arg for arg in args)
    od = var_select(colnames, *evaluated)

    missing_groups, group_keys = _select_group_renames(od, __data.group_by)

    if missing_groups:
        _warn_missing(missing_groups)

    final_od = {**{k: None for k in missing_groups}, **od}

    col_list = []
    for k,v in final_od.items():
        col = columns[k]
        col_list.append(col if v is None else col.label(v))

    return __data.append_op(
        _sql_with_only_columns(last_sel, col_list),
        group_by = group_keys
    )


@rename.register(LazyTbl)
def _rename(__data, **kwargs):
    sel = __data.last_select
    columns = lift_inner_cols(sel)

    # old_keys uses dict as ordered set
    old_to_new = {simple_varname(v):k for k,v in kwargs.items()}
    
    if None in old_to_new:
        raise KeyError("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    labs = [c.label(old_to_new[k]) if k in old_to_new else c for k,c in columns.items()]

    new_sel = _sql_with_only_columns(sel, labs)

    missing_groups, group_keys = _select_group_renames(old_to_new, __data.group_by)

    return __data.append_op(new_sel, group_by=group_keys)





from functools import singledispatch

from .calls import Call, BinaryOp, BinaryRightOp, MetaArg, UnaryOp, SliceOp, FuncArg
from .format import Formatter

# Symbolic
# =============================================================================

def create_binary_op(op_name, left_op = True):
    def _binary_op(self, x):
        if left_op:
            node = BinaryOp(op_name, strip_symbolic(self), strip_symbolic(x))
        else:
            node = BinaryRightOp(op_name, strip_symbolic(self), strip_symbolic(x))

        return self.__class__(node, ready_to_call = True)
    return _binary_op

def create_unary_op(op_name):
    def _unary_op(self):
        node = UnaryOp(op_name, strip_symbolic(self))

        return self.__class__(node, ready_to_call = True)

    return _unary_op


class Symbolic(object):
    def __init__(self, source = None, ready_to_call = False):
        self.__source = MetaArg("_") if source is None else source
        self.__ready_to_call = ready_to_call

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Handle numpy universal functions. E.g. np.sqrt(_)."""
        return array_ufunc(self, ufunc, method, *inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        return array_function(self, func, types, *args, **kwargs)

    # since we override __eq__, we must explicitly set the hash method back to default
    __hash__ = object.__hash__

    # allowed methods ----

    def __getattr__(self, x):
        # temporary hack working around ipython pretty.py printing
        #if x == "__class__": return Symbolic

        return Symbolic(BinaryOp(
                "__getattr__",
                self.__source,
                strip_symbolic(x)
                ))
                

    def __call__(self, *args, **kwargs) -> "Symbolic":
        if self.__ready_to_call:
            return self.__source(*args, **kwargs)

        return create_sym_call(self.__source, *args, **kwargs)

    def __getitem__(self, x):
        return Symbolic(BinaryOp(
                "__getitem__",
                self.__source,
                slice_to_call(x),
                ),
                ready_to_call = True)

    
    def __invert__(self):
        if isinstance(self.__source, Call) and self.__source.func == "__invert__":
            return self.__source.args[0]
        else: 
            return self.__op_invert()


    def __op_invert(self):
        return Symbolic(UnaryOp('__invert__', self.__source), ready_to_call = True)

    def __rshift__(self, x):

        # Note that this and __rrshift__ are copied from Call
        stripped = strip_symbolic(x)

        if isinstance(stripped, Call):
            lhs_call = self.__source
            return self.__class__(Call._construct_pipe(lhs_call, stripped))
        # strip_symbolic(self)(x)
        # x is a symbolic
        raise NotImplementedError("Symbolic may only be used on right-hand side of >> operator.")

    def __rrshift__(self, x):
        if isinstance(x, (Symbolic, Call)):
            raise NotImplementedError()

        return strip_symbolic(self)(x)


    # banned methods ----

    __contains__ = None
    __iter__ = None

    def __bool__(self):
        raise TypeError("Symbolic objects can not be converted to True/False, or used "
                        "with these keywords: not, and, or.")


    # representation ----
        
    def __repr__(self):
        return Formatter().format(self.__source)

    # unary operators ----
    # note that __invert__ is handled in a custom way above
    __neg__ = create_unary_op("__neg__")
    __pos__ = create_unary_op("__pos__")
    __abs__ = create_unary_op("__abs__")


    # binary operators ----
    __add__ = create_binary_op("__add__")
    __sub__ = create_binary_op("__sub__")
    __mul__ = create_binary_op("__mul__")
    __matmul__ = create_binary_op("__matmul__")
    __truediv__ = create_binary_op("__truediv__")
    __floordiv__ = create_binary_op("__floordiv__")
    __mod__ = create_binary_op("__mod__")
    __divmod__ = create_binary_op("__divmod__")
    __pow__ = create_binary_op("__pow__")
    __lshift__ = create_binary_op("__lshift__")
    __and__ = create_binary_op("__and__")
    __xor__ = create_binary_op("__xor__")
    __or__ = create_binary_op("__or__")
    __gt__ = create_binary_op("__gt__")
    __lt__ = create_binary_op("__lt__")
    __eq__ = create_binary_op("__eq__")
    __ne__ = create_binary_op("__ne__")
    __ge__ = create_binary_op("__ge__")
    __le__ = create_binary_op("__le__")


    __radd__ = create_binary_op("__radd__", False)
    __rsub__ = create_binary_op("__rsub__", False)
    __rmul__ = create_binary_op("__rmul__", False)
    __rmatmul__ = create_binary_op("__rmatmul__", False)
    __rtruediv__ = create_binary_op("__rtruediv__", False)
    __rfloordiv__ = create_binary_op("__rfloordiv__", False)
    __rmod__ = create_binary_op("__rmod__", False)
    __rdivmod__ = create_binary_op("__rdivmod__", False)
    __rpow__ = create_binary_op("__rpow__", False)
    __rlshift__ = create_binary_op("__rlshift__", False)
    __rand__ = create_binary_op("__rand__", False)
    __rxor__ = create_binary_op("__rxor__", False)
    __ror__ = create_binary_op("__ror__", False)


def create_sym_call(__source, *args, **kwargs):
    return Symbolic(Call(
            "__call__",
            strip_symbolic(__source),
            *map(strip_symbolic, args),
            **{k: strip_symbolic(v) for k,v in kwargs.items()}
            ),
            ready_to_call = True)


def slice_to_call(x):

    # TODO: uses similar code to SliceOp. make a walk_slice function?
    def f_strip(s):
        if isinstance(s, slice):
            return slice(*map(strip_symbolic, (s.start, s.stop, s.step)))

        return strip_symbolic(s)


    if isinstance(x, tuple):
        arg = tuple(map(f_strip, x))
    else:
        arg = f_strip(x)

    return SliceOp("__siu_slice__", arg)


def strip_symbolic(x):
    if isinstance(x, Symbolic):
        return x.__dict__["_Symbolic__source"]

    return x


def explain(symbol):
    """Print representation that resembles code used to create symbol."""
    if isinstance(symbol, Symbolic):
        return str(strip_symbolic(symbol))

    return str(symbol)


# Special numpy ufunc dispatcher
# =============================================================================
# note that this is essentially what dispatchers.symbolic_dispatch does...
# details on numpy array dispatch: https://github.com/numpy/numpy/issues/21387

@singledispatch
def array_function(self, func, types, *args, **kwargs):
    return func(*args, **kwargs)


@array_function.register(Call)
def _array_function_call(self, func, types, *args, **kwargs):
    return Call("__call__", FuncArg(array_function), self, func, types, *args, **kwargs)


@array_function.register(Symbolic)
def _array_function_sym(self, func, types, *args, **kwargs):
    f_concrete = array_function.dispatch(Call)

    call = f_concrete(
        strip_symbolic(self),
        func,
        types, 
        *map(strip_symbolic, args),
        **{k: strip_symbolic(v) for k, v in kwargs.items()}
    )

    return Symbolic(call)


@singledispatch
def array_ufunc(self, ufunc, method, *inputs, **kwargs):
    return getattr(ufunc, method)(*inputs, **kwargs)

@array_ufunc.register(Call)
def _array_ufunc_call(self, ufunc, method, *inputs, **kwargs):

    return Call("__call__", FuncArg(array_ufunc), self, ufunc, method, *inputs, **kwargs)


@array_ufunc.register(Symbolic)
def _array_ufunc_sym(self, ufunc, method, *inputs, **kwargs):
    f_concrete = array_ufunc.dispatch(Call)

    call = f_concrete(
        strip_symbolic(self),
        ufunc,
        method, 
        *map(strip_symbolic, inputs),
        **{k: strip_symbolic(v) for k, v in kwargs.items()}
    )

    return Symbolic(call)


# Do some gnarly method setting on Symbolic -----------------------------------
# =============================================================================




from siuba.sql.utils import get_dialect_translator, mock_sqlalchemy_engine
from siuba.dply.verbs import collect
from siuba.sql import LazyTbl
import pytest

@pytest.mark.parametrize('name', [
    'redshift',
    'postgresql',
    'sqlite'
    ])
def test_get_dialect_translator(name):
    get_dialect_translator(name)

def test_mock_sqlalchemy_engine_dialect():
    engine = mock_sqlalchemy_engine("postgresql")
    assert engine.dialect.name == "postgresql"

    engine = mock_sqlalchemy_engine("sqlite")
    assert engine.dialect.name == "sqlite"

def test_mock_sqlalchemy_engine_no_collect():
    engine = mock_sqlalchemy_engine("sqlite")
    tbl = LazyTbl(engine, "some_table", ["x"])
    assert collect(tbl) is None



from siuba.dply.verbs import (
        simple_varname,
        mutate,
        transmute,
        )

from ..backend import LazyTbl, SqlLabelReplacer
from ..translate import ColumnCollection
from ..utils import (
    _sql_with_only_columns,
    lift_inner_cols
)

from sqlalchemy import sql
# TODO: currently needed for select, but can we remove pandas?

from siuba.dply.across import _require_across, _eval_with_context


@mutate.register(LazyTbl)
def _mutate(__data, *args, **kwargs):
    # TODO: verify it can follow a renaming select

    # track labeled columns in set
    if not (len(args) or len(kwargs)):
        return __data.append_op(__data.last_op)

    names, sel_out = _mutate_cols(__data, args, kwargs, "Mutate")
    return __data.append_op(sel_out)


def _sql_upsert_columns(sel, new_columns: "list[base.Label | base.Column]"):
    orig_cols = lift_inner_cols(sel)
    replaced = {**orig_cols}

    for new_col in new_columns:
        replaced[new_col.name] = new_col
    return _sql_with_only_columns(sel, list(replaced.values()))


def _select_mutate_result(src_sel, expr_result):
    dst_alias = src_sel.alias()
    src_columns = set(lift_inner_cols(src_sel))
    replacer = SqlLabelReplacer(set(src_columns), dst_alias.columns)

    if isinstance(expr_result, ColumnCollection):
        replaced_cols = list(map(replacer, expr_result))
        orig_cols = expr_result
    #elif isinstance(expr_result, None):
    #    pass
    else:
        replaced_cols = [replacer(expr_result)]
        orig_cols = [expr_result]

    if replacer.applied:
        return _sql_upsert_columns(dst_alias.select(), replaced_cols)

    return _sql_upsert_columns(src_sel, orig_cols)


def _eval_expr_arg(__data, sel, func, verb_name, window=True):
    inner_cols = lift_inner_cols(sel)

    # case 1: simple names ----
    simple_name = simple_varname(func)
    if simple_name is not None:
        return inner_cols[simple_name]

    # case 2: across ----
    _require_across(func, verb_name)

    cols_result = _eval_with_context(__data, window, inner_cols, func)

    # TODO: remove or raise a more informative error
    assert isinstance(cols_result, ColumnCollection), type(cols_result)

    return cols_result


def _eval_expr_kwarg(__data, sel, func, new_name, verb_name, window=True):
    inner_cols = lift_inner_cols(sel)

    expr_shaped = __data.shape_call(func, window, verb_name = verb_name, arg_name = new_name)
    new_col, windows, _ = __data.track_call_windows(expr_shaped, inner_cols)

    if isinstance(new_col, ColumnCollection):
        raise TypeError(
            f"{verb_name} named arguments must return a single column, but `{new_name}` "
            "returned multiple columns."
        )

    return new_col.label(new_name)


def _mutate_cols(__data, args, kwargs, verb_name):
    result_names = {}     # used as ordered set
    sel = __data.last_select

    for ii, func in enumerate(args):
        cols_result = _eval_expr_arg(__data, sel, func, verb_name)

        # replace any labels that require a subquery ----
        sel = _select_mutate_result(sel, cols_result)

        if isinstance(cols_result, ColumnCollection):
            result_names.update({k: True for k in cols_result.keys()})
        else:
            result_names[cols_result.name] = True

    
    for new_name, func in kwargs.items():
        labeled = _eval_expr_kwarg(__data, sel, func, new_name, verb_name)

        sel = _select_mutate_result(sel, labeled)
        result_names[new_name] = True


    return list(result_names), sel


@transmute.register(LazyTbl)
def _transmute(__data, *args, **kwargs):
    # will use mutate, then select some cols
    result_names, sel = _mutate_cols(__data, args, kwargs, "Transmute")

    # transmute keeps grouping cols, and any defined in kwargs
    missing = [x for x in __data.group_by if x not in result_names]
    cols_to_keep = [*missing, *result_names]

    columns = lift_inner_cols(sel)
    sel_stripped = _sql_with_only_columns(sel, [columns[k] for k in cols_to_keep])

    return __data.append_op(sel_stripped)



"""
This module holds default translations from pandas syntax to sql for 3 kinds of operations...

1. scalar - elementwise operations (e.g. array1 + array2)
2. aggregation - operations that result in a single number (e.g. array1.mean())
3. window - operations that do calculations across a window
            (e.g. array1.lag() or array1.expanding().mean())

It is broken into 5 sections:

* Warnings
* Column classes
* Custom over clauses
* Translation function generators (e.g. sql_agg("stdev"))
* SqlTranslator for translating symbolic expressions.

"""

from sqlalchemy import sql

from siuba.siu import FunctionLookupBound, FunctionLookupError


# warning for when sql defaults differ from pandas ============================
import warnings


class SqlFunctionLookupError(FunctionLookupError): pass


class SiubaSqlRuntimeWarning(UserWarning): pass

def warn_arg_default(func_name, arg_name, arg, correct):
    warnings.warn(
            "\n{func_name} sql translation defaults "
            "{arg_name} to {arg}. To return identical result as pandas, use "
            "{arg_name} = {correct}.\n\n"
            "This warning only displays once per function".format(
                func_name = func_name, arg_name = arg_name, arg = repr(arg), correct = repr(correct)
                ),
            SiubaSqlRuntimeWarning
            )


# Column data types ===========================================================

from sqlalchemy.sql.elements import ClauseElement

class SqlBase(ClauseElement): pass

class SqlColumn(SqlBase): pass

class SqlColumnAgg(SqlBase): pass


# Columns container ===========================================================

from sqlalchemy.sql import ColumnCollection

# Custom over clause handling  ================================================

from sqlalchemy.sql.elements import Over


class CustomOverClause(Over):
    """Base class for custom window clauses in SQL translation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def set_over(self, group_by, order_by):
        raise NotImplementedError()

    def has_over(self):
        return self.order_by is not None or self.group_by is not None


    @classmethod
    def func(cls, name):
        raise NotImplementedError()


    


class AggOver(CustomOverClause):
    """Over clause for uses of functions min, max, avg, that return one value.

    Note that this class does not set order by, which is how these functions
    generally become their cumulative versions.

    E.g. mean(x) -> AVG(x) OVER (partition_by <group vars>)
    """

    def set_over(self, group_by, order_by = None):
        self.partition_by = group_by
        return self

    @classmethod
    def func(cls, name):
        sa_func = getattr(sql.func, name)
        def f(codata, col, *args, **kwargs) -> AggOver:
            return cls(sa_func(col, *args, **kwargs))

        return f


class RankOver(CustomOverClause): 
    """Over clause for ranking functions.

    Note that in python we might call rank(col), but in SQL the ranking column
    is defined using order by.

    E.g. rank(y) -> rank() OVER (partition by <group vars> order by y)
    """
    def set_over(self, group_by, order_by = None):
        crnt_partition = getattr(self.partition_by, 'clauses', tuple())
        self.partition_by = sql.elements.ClauseList(*crnt_partition, *group_by.clauses)
        return self

    @classmethod
    def func(cls, name):
        sa_func = getattr(sql.func, name)
        def f(codata, col) -> RankOver:
            return cls(sa_func(), order_by = col)

        return f


class CumlOver(CustomOverClause):
    """Over clause for cumulative versions of functions like sum, min, max.

    Note that this class is also currently used for aggregates that might require
    ordering, like nth, first, etc..

    e.g. cumsum(x) -> SUM(x) OVER (partition by <group vars> order by <order vars>)
    e.g. nth(0) -> NTH_VALUE(1) OVER (partition by <group vars> order by <order vars>)

    """
    def set_over(self, group_by, order_by):
        self.partition_by = group_by


        # do not override order by if it was set by the user. this might happen
        # in functions like nth, which gives the option to set it.
        if self.order_by is None or not len(self.order_by):
            if not len(order_by):
                warnings.warn(
                        "No order by columns explicitly set in window function. SQL engine"
                        "does not guarantee a row ordering. Recommend using an arrange beforehand.",
                        RuntimeWarning
                        )

            self.order_by = order_by


        return self

    @classmethod
    def func(cls, name, rows=(None, 0)):
        sa_func = getattr(sql.func, name)
        def f(codata, col, *args, **kwargs) -> CumlOver:
            return cls(sa_func(col, *args, **kwargs), rows = rows)

        return f

# convenience aliases for class methods above
# TODO MC-NOTE: funcs like AggOver.func should not have codata as the first argument, 
# since they are simple sqlalchemy subclasses. However, they're used as function
# factories in the dialects...
win_agg = AggOver.func
win_over = RankOver.func
win_cumul = CumlOver.func


# Simple clause translation generators ========================================

def sql_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda codata, col: sa_func(col)

def sql_scalar(name):
    sa_func = getattr(sql.func, name)
    return lambda codata, col, *args: sa_func(col, *args)

def sql_colmeth(meth, *outerargs):
    def f(codata, col, *args) -> SqlColumn:
        return getattr(col, meth)(*outerargs, *args)
    return f

def sql_ordered_set(name, is_analytic=False):
    """Generate function for ordered and hypothetical set aggregates.

    Hypothetical-set aggregates take an argument, and return a value for each
    element of the argument. For example: rank(2) WITHIN GROUP (order by x).
    In this case, the hypothetical ranks 2 relative to x.

    Ordered set aggregates are like percentil_cont(.5) WITHIN GROUP (order by x),
    which calculates the median of x.
    """
    sa_func = getattr(sql.func, name)

    if is_analytic:
        return lambda codata, col, *args: AggOver(
            sa_func(*args).within_group(col)
        )

    return lambda codata, col, *args: sa_func(*args).within_group(col)

# Handling not implemented translations ----

def sql_not_impl(msg = ""):

    return FunctionLookupBound(msg or "function not implemented")

def win_absent(name):
    # Return an error, that is picked up by the translator.
    # this allows us to report errors at translation, rather than call time.
    return FunctionLookupBound("SQL dialect does not support window function {}.".format(name))

# Annotations ----
# these functions wrap translation generators, in order to provide metadata
# for running, e.g., unit tests. This is important in cases where the SQL 
# output does not match exactly what pandas does (for example, it returns a 
# float, when pandas returns an int).

def annotate(f = None, **kwargs):
    # allow it to work as a decorator
    if f is None:
        return lambda f: annotate(f, **kwargs)

    if hasattr(f, "operation"):
        raise ValueError("function already has an operation attribute")

    f.operation = kwargs

    return f

def wrap_annotate(f, **kwargs):
    from functools import wraps

    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    wrapper.operation = kwargs

    return wrapper


#  Translator =================================================================

from siuba.ops.translate import create_pandas_translator
from functools import singledispatch


def extend_base(cls, **kwargs):
    """Register concrete methods onto generic functions for pandas Series methods."""
    from siuba.ops import ALL_OPS
    for meth_name, f in kwargs.items():
        ALL_OPS[meth_name].register(cls, f)


# TODO: should inherit from a ITranslate class (w/ abstract translate method)
class SqlTranslator:
    """Translates symbolic column operations to sqlalchemy clauses.

    Note that SqlTranslator's main job is to hold two actual translators--
    one for windowed and one for aggregation contexts--and to call the
    translate method to the correct one.


    Example::

       from siuba.sql.translate import (
           SqlTranslator,
           SqlColumn, SqlColumnAgg,
           AggOver, sql_not_impl
           )

       base   = {"__add__": sql_colmeth("add")}
       window = {"mean":    AggOver.func("mean")}
       agg    = {"mean":    sql_not_impl("group by mean")}

       trans = SqlTranslator.from_mappings(
               base, window, agg,
               SqlColumn, SqlColumnAgg
               )

       from siuba import _
       trans.translate(_ + _.mean())
       trans.translate(_ + _.mean(), window = False)
   """

    def __init__(self, window, aggregate):
        self.window = window
        self.aggregate = aggregate

    def translate(self, expr, window = True):
        """Convert an AST of method chains to an AST of function calls."""

        if window:
            return self.window.translate(expr)

        return self.aggregate.translate(expr)


    def shape_call(
            self,
            call, window = True, str_accessors = False,
            verb_name = None, arg_name = None,
            ):
        """Return a siu Call that creates dialect specific SQL when called."""

        from siuba.siu import Call, MetaArg, strip_symbolic, Lazy, str_to_getitem_call
        from siuba.siu.visitors import CodataVisitor

        call = strip_symbolic(call)

        if isinstance(call, Call):
            pass
        elif str_accessors and isinstance(call, str):
            # verbs that can use strings as accessors, like group_by, or
            # arrange, need to convert those strings into a getitem call
            return str_to_getitem_call(call)
        elif isinstance(call, (sql.elements.ClauseElement)):
            return Lazy(call)
        elif callable(call):
            #TODO: should not happen here
            return Call("__call__", call, MetaArg('_'))

        else:
            # verbs that use literal strings, need to convert them to a call
            # that returns a sqlalchemy "literal" object
            _lit = convert_literal(self.window.dispatch_cls(), call)
            return Lazy(_lit)

        # raise informative error message if missing translation
        try:
            # TODO: MC-NOTE -- scaffolding in to verify prior behavior works
            shaped_call = self.translate(call, window = window)
            if window:
                trans = self.window
            else:
                trans = self.aggregate

            # TODO: MC-NOTE - once all sql singledispatch funcs are annotated
            # with return types, then switch object back out
            # alternatively, could register a bounding class, and remove
            # the result type check
            v = CodataVisitor(trans.dispatch_cls, object)
            return v.visit(shaped_call)
            
        except FunctionLookupError as err:
            raise SqlFunctionLookupError.from_verb(
                    verb_name or "Unknown",
                    arg_name or "Unknown",
                    err,
                    short = True
                    )


    def from_mappings(WinCls, AggCls):
        from siuba.ops import ALL_OPS

        return SqlTranslator(
                window = create_pandas_translator(ALL_OPS, WinCls, sql.elements.ClauseElement),
                aggregate = create_pandas_translator(ALL_OPS, AggCls, sql.elements.ClauseElement)
                )


@singledispatch
def convert_literal(codata, lit):
    return sql.literal(lit)



"""
Implements LazyTbl to represent tables of SQL data, and registers it on verbs.

This module is responsible for the handling of the "table" side of things, while
translate.py handles translating column operations.


"""

from ..backend import LazyTbl
from ..utils import _sql_simplify_select

from siuba.dply.verbs import show_query


@show_query.register(LazyTbl)
def _show_query(tbl, simplify = False, return_table = True):
    #query = tbl.last_op #if not simplify else 
    compile_query = lambda query: query.compile(
                dialect = tbl.source.dialect,
                compile_kwargs = {"literal_binds": True}
            )


    if simplify:
        # try to strip table names and labels where unnecessary
        simple_sel = _sql_simplify_select(tbl.last_select)

        explained = compile_query(simple_sel)
    else:
        # use a much more verbose query
        explained = compile_query(tbl.last_select)

    if return_table:
        print(str(explained))
        return tbl

    return str(explained)



from .calls import ALL_OPS, Call, MetaArg

class Formatter:
    def __init__(self): pass

    def format(self, call, pad = 0):
        """Return a Symbolic or Call back as a nice tree, with boxes for nodes."""

        fmt_block = "█─"
        fmt_pipe = "├─"
        
        if isinstance(call, MetaArg):
            return "_"

        if isinstance(call, Call):
            call_str = fmt_block + ALL_OPS.get(call.func, repr(call.func))

            args_str = [self.format(arg) for arg in call.args]

            # format keyword args, making sure "└─<key> = █─" aligns box's children
            kwargs_str = []
            for k, v in call.kwargs.items():
                kwargs_str.append(
                        k + " = " + self.format(v, pad = len(k) + 3)
                        )

            all_args = [*args_str, *kwargs_str]
            padded = []
            for ii, entry in enumerate(all_args):
                chunk = self.fmt_pipe(
                        entry,
                        final = ii == len(all_args) - 1,
                        pad = pad
                        )
                padded.append(chunk)

            return "".join([call_str, *padded])

        return repr(call)


    @staticmethod
    def fmt_pipe(x, final = False, pad = 0):
        if not final:
            connector = "│ " if not final else "  "
            prefix = "├─"
        else:
            connector = "  "
            prefix = "└─"

        connector = "\n" + " "*pad + connector
        prefix = "\n" + " "*pad + prefix
        # NOTE: because visiting is depth first, this is essentially prepending
        # the text to the left.
        return prefix + connector.join(x.splitlines())






import numpy as np
import pandas as pd
import pytest

from pandas.testing import assert_frame_equal, assert_series_equal

from siuba.dply.verbs import collect
from siuba.siu import Symbolic
from siuba.tests.helpers import data_frame
from . import pivot_wider, pivot_wider_spec, build_wider_spec

from .test_pivot import assert_equal_query2, assert_series_equal2

_ = Symbolic()

def test_pivot_all_cols(backend):
    # TODO: SQL - bigquery distinct not ordered, so col order is out
    src = backend.load_df(
        data_frame(key = ["x", "y", "z"], val = [1, 2, 3])
    )
    dst = data_frame(x = 1, y = 2, z = 3)

    pv = pivot_wider(src, names_from=_.key, values_from=_.val)

    # Note: duckdb is ordered
    assert_equal_query2(pv, dst, sql_kwargs = {"check_like": True}, sql_ordered=True)


def test_pivot_id_cols_default_preserve(backend):
    src = backend.load_df(
        data_frame(a = 1, key = ["x", "y"], val = [1, 2])
    )
    dst = data_frame(a = [1], x = [1], y = [2])

    pv = pivot_wider(src, names_from = _.key, values_from = _.val)

    # Note: duckdb is ordered
    assert_equal_query2(pv, dst, sql_kwargs = {"check_like": True}, sql_ordered=True)


def test_pivot_implicit_missings_to_explicit():
    src = data_frame(a = [1, 2], key = ["x", "y"], val = [1, 2])
    dst = data_frame(a = [1, 2], x = [1., None], y = [None, 2.])

    pv = pivot_wider(src, names_from = _.key, values_from = _.val)

    assert_equal_query2(pv, dst, sql_ordered=True)


def test_pivot_implicit_missings_to_explicit_from_spec(backend):
    df = data_frame(a = [1, 2], key = ["x", "y"], val = [1, 2])
    dst = data_frame(a = [1, 2], x = [1., None], y = [None, 2.])

    sp = pd.DataFrame({".name": ["x", "y"], ".value": ["val"]*2, "key": ["x", "y"]})

    src = backend.load_df(df)
    pv = pivot_wider_spec(src, sp)

    assert_equal_query2(
        pv,
        dst,
        check_dtype=False,
        # Note: duckdb is ordered
        sql_ordered=False,
    )


@pytest.mark.skip_backend("bigquery")
def test_error_overwriting_existing_column(skip_backend, backend):
    # TODO: SQL - bigquery distinct not ordered, so col order is out
    src = backend.load_df(
        data_frame(a = [1, 1], key = ["a", "b"], val = [1, 2])
    )
    
    # a bunch of snapshot tests...
    # should error
    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, names_from = _.key, values_from = _.val)

    assert "1 duplicate name(s)" in exc_info.value.args[0]

    pv = collect(
        pivot_wider(src, names_from = _.key, values_from = _.val, names_repair = "unique")
    )

    # TODO: note that our names start with 0-based indexing
    assert list(pv.columns) == ["a___0", "a___1", "b"]


@pytest.mark.skip_backend("bigquery")
def test_names_repair_unique(skip_backend, backend):
    src = backend.load_df(
        data_frame(test = ["a", "b"], name = ["test", "test2"], value = [1, 2])
    )
    dst = data_frame(test_0 = ["a", "b"], test_1 = [1., None], test2_2 = [None, 2.])

    pv = pivot_wider(src, names_repair=lambda x: [f"{v}_{ii}" for ii, v in enumerate(x)])

    assert_equal_query2(pv, dst, sql_ordered=True)


def test_names_repair_minimal():
    src = data_frame(test = ["a", "b"], name = ["test", "test2"], value = [1, 2])
    dst = pd.DataFrame(
        [
            ["a", 1., None],
            ["b", None, 2.]
        ],
        columns = ["test", "test", "test2"]
    )

    pv = pivot_wider(src, names_repair = "minimal")

    assert_frame_equal(pv, dst)


def test_grouping_preserved():
    gdf = data_frame(g = 1, k = "x", v = 2).groupby("g")

    pv = pivot_wider(gdf, names_from = _.k, values_from = _.v)

    assert isinstance(pv, pd.core.groupby.DataFrameGroupBy)


@pytest.mark.skip_backend("bigquery")
def test_weird_column_name_select(skip_backend, backend):
    src = backend.load_df(
        pd.DataFrame({"...8": ["x", "y", "z"], "val": [1, 2, 3]})
    )
    dst = data_frame(x = 1, y = 2, z = 3)

    pv = pivot_wider(src, names_from = _["...8"], values_from = _.val)

    assert_equal_query2(pv, dst, sql_kwargs = {"check_like": True}, sql_ordered=True)


@pytest.mark.skip("Won't do")
def test_data_frame_columns_pivot_correctly():
    # this concept doesn't exist in pandas. here is the R code for frame construction.
    # note that the columns x and y belong to a side-by-side data.frame
    # df <- tibble(
    #   i = c(1, 2, 1, 2),
    #   g = c("a", "a", "b", "b"),
    #   d = tibble(x = 1:4, y = 5:8)
    # )
    pass


def test_names_from_required_error(backend):
    # column name doesn't exist
    src = backend.load_df(
        data_frame(key = "x", val = 1)
    )

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, values_from = _.val)

    assert "name" in exc_info.value.args[0]


def test_values_from_required_error(backend):
    src = backend.load_df(
        data_frame(key = "x", val = 1)
    )

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, names_from = _.key)

    assert "value" in exc_info.value.args[0]


def test_names_from_no_match_error(backend):
    src = backend.load_df(
        data_frame(key = "x", val = 1)
    )

    with pytest.raises(ValueError) as exc_info:
        pv = pivot_wider(src, names_from = _.startswith("foo"), values_from = _.val)

    assert "`names_from` must" in exc_info.value.args[0]


def test_values_from_no_match_error(backend):
    src = backend.load_df(
        data_frame(key = "x", val = 1)
    )

    with pytest.raises(ValueError) as exc_info:
        pv = pivot_wider(src, names_from = _.key, values_from = _.startswith("foo"))

    assert "`values_from` must" in exc_info.value.args[0]


def test_values_fn_informative_error_pandas_fastpath():
    src = data_frame(name = ["a", "a"], value = [1, 2])

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, values_fn = {"value": "cumsum"})

    assert "bug in pandas" in exc_info.value.args[0]
    assert "cumsum" in exc_info.value.args[0]


def test_values_fn_informative_error_lambda():
    src = data_frame(name = ["a", "a"], value = [1, 2])

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, values_fn = {"value": lambda x: x.cumsum()})

    assert "Must produce aggregated value" in exc_info.value.args[0]


@pytest.mark.skip("TODO: pivot from spec")
def test_pivot_manual_spec_with_columns_that_dont_identify_rows():
    ## Looking for `x = 1L`
    #spec <- tibble(.name = "name", .value = "value", x = 1L)
    #
    ## But that doesn't exist here...
    #df <- tibble(key = "a", value = 1L, x = 2L)
    #expect_identical(
    #  pivot_wider_spec(df, spec, id_cols = key),
    #  tibble(key = "a", name = NA_integer_)
    #)
    #
    ## ...or here
    #df <- tibble(key = character(), value = integer(), x = integer())
    #expect_identical(
    #  pivot_wider_spec(df, spec, id_cols = key),
    #  tibble(key = character(), name = integer())
    #)
    #})
    pass


def test_pivot_manual_spec_zero_rows():
    spec = pd.DataFrame({".name": ["name"], ".value": "value", "x": 1})
    src = data_frame(value = [], x = [])

    assert_frame_equal(
        pivot_wider_spec(src, spec),
        pd.DataFrame({"name": []}, dtype = "object")
    )


def test_names_expand_basic():
    """can use `names_expand` to get sorted and expanded column names (#770)"""

    name1 = pd.Categorical([None, "x"], ["x", "y"])
    src = data_frame(name1 = name1, name2 = ["c", "d"], value = [1, 2])

    assert_frame_equal(
        pivot_wider(src, names_from = _[_.name1, _.name2], names_expand=True),
        data_frame(x_c = None, x_d = 2, y_c = None, y_d = None, nan_c = 1, nan_d = None)
    )


def test_names_expand_fills_implicit_missings():
    """can fill only implicit missings from `names_expand`"""
    name1 = pd.Categorical([None, "x"], ["x", "y"])
    src = data_frame(name1 = name1, name2 = ["c", "d"], value = [1, None])

    pv = pivot_wider(
        src,
        names_from = _[_.name1, _.name2],
        names_expand = True,
        values_fill = 0
    )

    assert_frame_equal(
        pv,
        data_frame(x_c = 0, x_d = np.nan, y_c = 0, y_d = 0, nan_c = 1., nan_d = 0)
    )


def test_id_expand_and_names_expand_works_with_zero_row_frames():
    src = data_frame(
        id = pd.Categorical([], ["b", "a"]),
        name = pd.Categorical([], ["a", "b"]),
        value = np.array([], dtype="int")
    )

    pv = pivot_wider(src, names_expand=True, id_expand=True)

    assert_frame_equal(
        pv,
        data_frame(
            id=["b", "a"],        # TODO: should preserve categorical
            a=[None, None],       # TODO: surprising these are not float64?
            b=[None, None]
        )
    )


# Column names ================================================================

@pytest.mark.xfail
def test_names_glue_basic(backend):
    # TODO: switch to use build_spec
    # TODO: prefix internal spec names with .
    src = backend.load_df(
        data_frame(x = ["X", "Y"], y = [1, 2], a = [1, 2], b = [1, 2])
    )

    pv = collect(
        pivot_wider(src, names_from=_[_.x, _.y], values_from=_[_.a, _.b], names_glue="{x}{y}_{.value}")
    )
    assert list(pv.columns) == ["X1_a", "Y2_a", "X1_b", "Y2_b"]
 

def test_names_sort_basic():
    # TODO: use spec
    from siuba.dply.forcats import fct_inorder
    src = data_frame(
        num = [1, 3, 2],
        fac = pd.Categorical(["Mon", "Wed", "Tue"], ["Mon", "Tues", "Wed"]),
        value = ["a", "b", "c"],
    )

    # should have cols ordered by position in the week
    pivot_wider(names_from = _[_.num, _.fac], names_sort = True)


def test_names_vary_slowest():
    src = data_frame(name = ["name1", "name2"], value1 = [1, 2], value2 = [4, 5])

    kwargs = {"names_from": _.name, "values_from": _[_.value1, _.value2]}

    pv_fastest = pivot_wider(src, **kwargs)
    assert (
        pv_fastest.columns.tolist() ==
            ["value1_name1", "value1_name2", "value2_name1", "value2_name2"]
    )

    pv_slowest = pivot_wider(src, names_vary="slowest", **kwargs)
    assert (
        pv_slowest.columns.tolist() ==
            ["value1_name1", "value2_name1", "value1_name2", "value2_name2"]
    )


def test_names_vary_is_validated():
    src = data_frame(name = ["a", "b"], value = [1, 2])

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, names_vary=1)

    assert "received argument: 1" in exc_info.value.args[0]


    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, names_vary="x")

    assert "received argument: 'x'" in exc_info.value.args[0]


def test_names_expand_always_sorts_columns():
    # TODO(ask): should we turn null values into "NA" in names?
    src = data_frame(
        name1 = pd.Categorical([None, "x"], ["x", "y"]),
        name2 = ["c", "d"],
        value = [1, 2]
    )

    pv = pivot_wider(src, names_from = _[_.name1, _.name2], names_expand=True)
    assert_frame_equal(
        pv,
        data_frame(x_c = None, x_d = 2, y_c = None, y_d = None, nan_c = 1, nan_d = None)
    )


@pytest.mark.xfail
def test_names_expand_always_sorts_column_names():
    """`names_expand` generates sorted column names even if no expansion is done"""
    # TODO: tidyr converts .name to string?

    src = data_frame(name = [2, 1], value = [1, 2])
    spec = build_wider_spec(src, names_expand = True)
    assert list(spec[".name"]) == ["1", "2"]


def test_names_expand_uses_all_categorical_levels():
    src = data_frame(name1 = ["a", "b"], name2 = ["c", "d"], value = [1, 2])
    spec = build_wider_spec(src, names_from = _[_.name1, _.name2], names_expand = True)

    assert list(spec[".name"]) == ["a_c", "a_d", "b_c", "b_d"]


def test_names_expand_is_validated():
    # TODO: test messages
    # TODO: use build spec
    src = data_frame(name = ["a", "b"], value = [1, 2])

    with pytest.raises(TypeError):
        pivot_wider(src, names_expand = 1)

    with pytest.raises(TypeError):
        pivot_wider(src, names_expand = "x")


# keys ========================================================================
# 11 tests

@pytest.mark.xfail
def test_pivot_can_override_default_keys():
    # TODO: pandas automatically orders the id columns when unstacking. Can we 
    # work around this?
    src = data_frame(
        row = [1, 2, 3],
        name = ["Sam", "Sam", "Bob"],
        var = ["age", "height", "age"],
        value = [10, 1.5, 20]
    )

    dst = data_frame(
        name = ["Sam", "Bob"],
        age = [10, 20],
        height = [1.5, None]
    )

    pv = pivot_wider(src, id_cols = _.name, names_from = _.var, values_from = _.value)

    assert_frame_equal(pv, dst)


def test_selecting_all_id_cols_excludes_names_from_values_from(backend):
    src = backend.load_df(
        data_frame(key = "x", name = "a", value = 1)
    )
    dst = data_frame(key = "x", a = 1)
    pv = pivot_wider(src, _[:])

    assert_equal_query2(pv, dst, sql_ordered=True)

    # TODO: also test pivot_wider_spec


def test_id_cols_overlaps_other_vars_error(backend):
    src = backend.load_df(
        data_frame(name = ["x", "y"], value = [1, 2])
    )

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, id_cols = _.name, names_from = _.name, values_from = _.value)

    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, id_cols = _.value, names_from = _.name, values_from = _.value)


def test_id_cols_no_column_match_error(backend):
    src = backend.load_df(
        data_frame(name = ["x", "y"], value = [1, 2])
    )
    with pytest.raises(ValueError) as exc_info:
        pivot_wider(src, id_cols = _.foo)

    msg = exc_info.value.args[0]
    assert "id_cols must" in msg
    assert "Could not find these columns: {'foo'}" in msg


def test_pivot_zero_row_frame_id_excludes_values_from(backend):
    src = backend.load_df(
        data_frame(key = pd.Series([], dtype="int"), name = [], value = [])
    )
    dst = data_frame(key = pd.Series([], dtype="int"))

    pv = pivot_wider(src, names_from = _.name, values_from = _.value)

    # SQL backends return a empty Index, pandas an empty RangeIndex ¯\_(ツ)_/¯
    assert_equal_query2(
        pv,
        dst,
        sql_kwargs = {"check_index_type": False, "check_dtype": False},
        sql_ordered=True
    )


#TODO:
#test_that("known bug - building a wider spec with a zero row data frame loses `values_from` info (#1249)", {


# id_expand ===============================================================-===

def test_id_expand_generates_sorted_rows():
    src = data_frame(id = [2, 1], name = ["a", "b"], value = [1, 2])
    pv = pivot_wider(src, id_expand = True)

    assert (pv.id == [1, 2]).all()


def test_id_expand_crosses_columns():
    src = data_frame(id1 = [1, 2], id2 = [3, 4], name = ["a", "b"], value = [1, 2])

    assert_frame_equal(
        pivot_wider(src, id_expand = True),
        data_frame(
            id1 = [1, 1, 2, 2],
            id2 = [3, 4, 3, 4],
            a = [1, None, None, None],
            b = [None, None, None, 2.0]
        )
    )


def test_id_expand_expands_all_levels_of_factor():
    id1 = pd.Categorical([None, "x"], ["x", "y"])
    src = data_frame(id1 = id1, id2 = [1, 2], name = ["a", "b"], value = [1, 2])

    pv = pivot_wider(src, id_expand=True)

    assert pv.id1.equals(pd.Series(["x", "x", "y", "y", None, None]))
    assert (pv.id2 == [1, 2, 1, 2, 1, 2]).all()


def test_id_expand_with_values_fill():
    """`id_expand` with `values_fill` only fills implicit missings"""

    id1 = pd.Categorical(["x", "x"], ["x", "y"])
    src = data_frame(id1 = id1, id2 = [1, 2], name = ["a", "b"], value = [1, None])

    pv = pivot_wider(src, id_expand = True, values_fill = 0)
    assert (pv.a == [1, 0, 0, 0]).all()
    assert pv.b.equals(pd.Series([0, None, 0, 0]))


@pytest.mark.xfail
def test_id_expand_values_fill_excludes_id_cols():
    # TODO: how does values_fill work?
    # TODO: it's possible we can't values_fill taking a dict, since we use pivot methods
    src = data_frame(id1 = ["none", "x"], id2 = [1, 2], name = ["a", "b"], value = [1, 2])

    pv = pivot_wider(src, id_expand=True, values_fill = list(id1 = 0))


def test_id_expand_arg_error():
    src = data_frame(name = ["a", "b"], value = [1, 2])

    with pytest.raises(TypeError):
        pivot_wider(src, id_expand = 1)

    with pytest.raises(TypeError):
        pivot_wider(src, id_expand = "x")


# non-unique keys =============================================================
# 8 tests

@pytest.mark.xfail
def test_duplciate_keys_produce_list_column_with_warning():
    # TODO: raises error, but this may be a wontdo
    src = data_frame(a = [1, 1, 2], key = ["x", "x", "x"], val = [1, 2, 3])
    pv = pivot_wider(src, names_from = _.key, values_from = _.val)

    assert False


@pytest.mark.xfail
def test_duplicated_key_warning_mentions_every_column():
    # TODO: similar to above -- wontdo?
    src = data_frame(key = ["x", "x"], a = [1, 2], b = [3, 4], c = [5, 6])

    pv = pivot_wider(src, names_from = _.key, values_from = _[_.a, _.b, _.c])

    assert False


@pytest.mark.xfail
def test_duplicated_key_warning_backticks_non_syntactic_names():
    # TODO: similar to above -- wontdo?
    assert False


def test_duplicate_error_resolved_by_values_fn():
    src = data_frame(a = [1, 1, 2], key = ["x", "x", "x"], val = [1, 2, 3])
    pv = pivot_wider(src, names_from = _.key, values_from = _.val, values_fn=list)

    assert_frame_equal(
        pv,
        data_frame(a = [1, 2], x = [[1, 2], [3]])
    )


def test_values_fn_arg_str(backend):
    src = backend.load_df(
        data_frame(a = [1, 1, 2], key = ["x", "x", "x"], val = [1, 2, 3])
    )
    pv = pivot_wider(src, names_from = _.key, values_from = _.val, values_fn="sum")

    assert_equal_query2(
        pv,
        data_frame(a = [1, 2], x = [3, 3]),
        sql_kwargs={"check_dtype": False}
    )


def test_values_fn_arg_lambda():
    f = lambda x: x.sum()
    src = data_frame(a = [1, 1, 2], key = ["x", "x", "x"], val = [1, 2, 3])
    pv = pivot_wider(src, names_from = _.key, values_from = _.val, values_fn=f)

    assert_frame_equal(
        pv,
        data_frame(a = [1, 2], x = [3, 3])
    )


def test_values_fn_on_no_duplicates():
    src = data_frame(a = [1, 2], key = ["x", "x"], val = [1, 2])
    pv = pivot_wider(src, names_from = _.key, values_from = _.val, values_fn=list)

    assert_frame_equal(
        pv,
        data_frame(a = [1, 2], x = [[1], [2]])
    )


# can fill missing cells ======================================================
# 3 tests
# TODO:

# multiple values =============================================================

def test_pivot_multiple_measure_cols():
    # TODO: use spec instead
    src = data_frame(row = 1, var = ["x", "y"], a = [1, 2], b = [3, 4])
    dst = data_frame(row = 1, a_x = 1, a_y = 2, b_x = 3, b_y = 4)

    pv = pivot_wider(src, names_from = _.var, values_from = _[_.a, _.b])

    assert_frame_equal(pv, dst)


def test_pivot_multiple_measure_no_id_cols():
    src = data_frame(var = ["x", "y"], a = [1, 2], b = [3, 4])
    dst = data_frame(a_x = 1, a_y = 2, b_x = 3, b_y = 4)

    pv = pivot_wider(src, names_from = _.var, values_from = _[_.a, _.b])

    assert_frame_equal(pv, dst)


@pytest.mark.skip("TODO: spec")
def test_pivot_column_order_matches_spec():
    """column order in output matches spec"""
    # TODO
    assert False

# unused_fn ===================================================================
# 6 tests
# TODO



import importlib

def test_data_imports():
    import siuba.data
    from siuba.data import __all__

    # note that we can't do import * inside a function, so programmatically fetch
    # each dataset
    for entry in __all__:
        getattr(siuba.data, entry)



from siuba.dply.verbs import distinct, mutate

from ..backend import LazyTbl, ordered_union
from ..utils import _sql_select, _sql_with_only_columns, lift_inner_cols

from .mutate import _mutate_cols


@distinct.register(LazyTbl)
def _distinct(__data, *args, _keep_all = False, **kwargs):
    if (args or kwargs) and _keep_all:
        raise NotImplementedError("Distinct with variables specified in sql requires _keep_all = False")
    
    result_names, inner_sel = _mutate_cols(__data, args, kwargs, "Distinct")

    # create list of final column names ----
    missing = [name for name in __data.group_by if name not in result_names]
    if not result_names:
        # use all columns if none passed to distinct
        all_names = list(lift_inner_cols(inner_sel).keys())
        final_names = ordered_union(missing, all_names)
    else:
        final_names = ordered_union(missing, result_names)


    if not (len(inner_sel._order_by_clause) or len(inner_sel._group_by_clause)):
        # select distinct has to include any columns in the order by clause,
        # so can only safely modify existing statement when there's no order by
        sel_cols = lift_inner_cols(inner_sel)
        distinct_cols = [sel_cols[k] for k in final_names]
        sel = _sql_with_only_columns(inner_sel, distinct_cols).distinct()
    else:
        # fallback to cte
        cte = inner_sel.alias()
        distinct_cols = [cte.columns[k] for k in final_names]
        sel = _sql_select(distinct_cols).select_from(cte).distinct()

    return __data.append_op(sel)



from .utils import operation, Namespace

# Infix operators -------------------------------------------------------------

ops_infix = Namespace(
    __add__ = operation('__add__', 'elwise', 2),
    __and__ = operation('__and__', 'elwise', 2),
    __eq__ = operation('__eq__', 'elwise', 2),
    __floordiv__ = operation('__floordiv__', 'elwise', 2),
    __ge__ = operation('__ge__', 'elwise', 2),
    __gt__ = operation('__gt__', 'elwise', 2),
    __invert__ = operation('__invert__', 'elwise', 1),
    __le__ = operation('__le__', 'elwise', 2),
    __lt__ = operation('__lt__', 'elwise', 2),
    __mod__ = operation('__mod__', 'elwise', 2),
    __mul__ = operation('__mul__', 'elwise', 2),
    __ne__ = operation('__ne__', 'elwise', 2),
    __neg__ = operation('__neg__', 'elwise', 1),
    __or__ = operation('__or__', 'elwise', 2),
    __pos__ = operation('__pos__', 'elwise', 1),
    __pow__ = operation('__pow__', 'elwise', 2),
    __radd__ = operation('__radd__', 'elwise', 2),
    __rand__ = operation('__rand__', 'elwise', 2),
    __rfloordiv__ = operation('__rfloordiv__', 'elwise', 2),
    __rmod__ = operation('__rmod__', 'elwise', 2),
    __rmul__ = operation('__rmul__', 'elwise', 2),
    __ror__ = operation('__ror__', 'elwise', 2),
    __round__ = operation('__round__', 'elwise', 1),
    __rpow__ = operation('__rpow__', 'elwise', 2),
    __rsub__ = operation('__rsub__', 'elwise', 2),
    __rtruediv__ = operation('__rtruediv__', 'elwise', 2),
    __rxor__ = operation('__rxor__', 'elwise', 2),
    __sub__ = operation('__sub__', 'elwise', 2),
    __truediv__ = operation('__truediv__', 'elwise', 2),
    __xor__ = operation('__xor__', 'elwise', 2),
    )

ops_infix_methods = Namespace(
    add           = operation('add', 'elwise', 2),
    div           = operation('div', 'elwise', 2),
    divide        = operation('divide', 'elwise', 2),
    divmod        = operation('divmod', None, 2),
    dot           = operation('dot', 'agg', 2),
    eq            = operation('eq', 'elwise', 2),
    floordiv      = operation('floordiv', 'elwise', 2),
    ge            = operation('ge', 'elwise', 2),
    gt            = operation('gt', 'elwise', 2),
    le            = operation('le', 'elwise', 2),
    lt            = operation('lt', 'elwise', 2),
    mod           = operation('mod', 'elwise', 2),
    mul           = operation('mul', 'elwise', 2),
    multiply      = operation('multiply', 'elwise', 2),
    ne            = operation('ne', 'elwise', 2),
    pow           = operation('pow', 'elwise', 2),
    radd          = operation('radd', 'elwise', 2),
    rdiv          = operation('rdiv', 'elwise', 2),
    rdivmod       = operation('rdivmod', None, 2),
    rfloordiv     = operation('rfloordiv', 'elwise', 2),
    rmod          = operation('rmod', 'elwise', 2),
    rmul          = operation('rmul', 'elwise', 2),
    round         = operation('round', 'elwise', 1),
    rpow          = operation('rpow', 'elwise', 2),
    rsub          = operation('rsub', 'elwise', 2),
    rtruediv      = operation('rtruediv', 'elwise', 2),
    sub           = operation('sub', 'elwise', 2),
    subtract      = operation('subtract', 'elwise', 2),
    truediv       = operation('truediv', 'elwise', 2),
    )


# Computation ---,

ops_compute = Namespace(
    abs     = operation('abs', 'elwise', 1),
    between = operation('between', 'elwise', 1),
    clip    = operation('clip', 'elwise', 1),
    isin = operation('isin', 'elwise', 1),

    all       = operation('all', 'agg', 1),
    any       = operation('any', 'agg', 1),
    corr      = operation('corr', 'agg', 2),
    count     = operation('count', 'agg', 1),
    cov       = operation('cov', 'agg', 2),
    is_unique = operation('is_unique', 'agg', 1, True, None),
    kurt      = operation('kurt', 'agg', 1),
    kurtosis  = operation('kurtosis', 'agg', 1),
    mad       = operation('mad', 'agg', 1),
    max       = operation('max', 'agg', 1),
    mean      = operation('mean', 'agg', 1),
    median    = operation('median', 'agg', 1),
    min       = operation('min', 'agg', 1),
    nunique   = operation('nunique', 'agg', 1),
    prod      = operation('prod', 'agg', 1),
    product   = operation('product', 'agg', 1),
    quantile  = operation('quantile', 'agg', 1),
    sem       = operation('sem', 'agg', 1),
    skew      = operation('skew', 'agg', 1),
    std       = operation('std', 'agg', 1),
    sum       = operation('sum', 'agg', 1),
    var       = operation('var', 'agg', 1),

    autocorr     = operation('autocorr', 'window', 1),
    cummax       = operation('cummax', 'window', 1),
    cummin       = operation('cummin', 'window', 1),
    cumprod      = operation('cumprod', 'window', 1),
    cumsum       = operation('cumsum', 'window', 1),
    diff         = operation('diff', 'window', 1),
    is_monotonic = operation('is_monotonic', 'window', 1, True, None),
    is_monotonic_decreasing = operation('is_monotonic_decreasing', 'window', 1, True, None),
    is_monotonic_increasing = operation('is_monotonic_increasing', 'window', 1, True, None),
    pct_change = operation('pct_change', 'window', 1),
    rank       = operation('rank', 'window', 1),

    compound     = operation('compound', None, 1),
    describe     = operation('describe', None, 1),
    factorize    = operation('factorize', None, 1),
    mode         = operation('mode', None, 1),
    nlargest     = operation('nlargest', None, 1),
    nsmallest    = operation('nsmallest', None, 1),
    unique       = operation('unique', None, 1),
    value_counts = operation('value_counts', None, 1),
    )


# Strings ---------------------------------------------------------------------

ops_str = Namespace(
    capitalize    = operation('capitalize', 'elwise', 1, False, 'str'),
    center        = operation('center', 'elwise', 1, False, 'str'),
    contains      = operation('contains', 'elwise', 1, False, 'str'),
    count         = operation('count', 'elwise', 1, False, 'str'),
    encode        = operation('encode', 'elwise', 1, False, 'str'),
    endswith      = operation('endswith', 'elwise', 1, False, 'str'),
    find          = operation('find', 'elwise', 1, False, 'str'),
    findall       = operation('findall', 'elwise', 1, False, 'str'),
    isalnum       = operation('isalnum', 'elwise', 1, False, 'str'),
    isalpha       = operation('isalpha', 'elwise', 1, False, 'str'),
    isdecimal     = operation('isdecimal', 'elwise', 1, False, 'str'),
    isdigit       = operation('isdigit', 'elwise', 1, False, 'str'),
    islower       = operation('islower', 'elwise', 1, False, 'str'),
    isnumeric     = operation('isnumeric', 'elwise', 1, False, 'str'),
    isspace       = operation('isspace', 'elwise', 1, False, 'str'),
    istitle       = operation('istitle', 'elwise', 1, False, 'str'),
    isupper       = operation('isupper', 'elwise', 1, False, 'str'),
    len           = operation('len', 'elwise', 1, False, 'str'),
    ljust         = operation('ljust', 'elwise', 1, False, 'str'),
    lower         = operation('lower', 'elwise', 1, False, 'str'),
    lstrip        = operation('lstrip', 'elwise', 1, False, 'str'),
    match         = operation('match', 'elwise', 1, False, 'str'),
    pad           = operation('pad', 'elwise', 1, False, 'str'),
    replace       = operation('replace', 'elwise', 1, False, 'str'),
    rfind         = operation('rfind', 'elwise', 1, False, 'str'),
    rjust         = operation('rjust', 'elwise', 1, False, 'str'),
    rsplit        = operation('rsplit', 'elwise', 1, False, 'str'),
    rstrip        = operation('rstrip', 'elwise', 1, False, 'str'),
    slice         = operation('slice', 'elwise', 1, False, 'str'),
    slice_replace = operation('slice_replace', 'elwise', 1, False, 'str'),
    split         = operation('split', 'elwise', 1, False, 'str'),
    startswith    = operation('startswith', 'elwise', 1, False, 'str'),
    strip         = operation('strip', 'elwise', 1, False, 'str'),
    swapcase      = operation('swapcase', 'elwise', 1, False, 'str'),
    title         = operation('title', 'elwise', 1, False, 'str'),
    upper         = operation('upper', 'elwise', 1, False, 'str'),
    wrap          = operation('wrap', 'elwise', 1, False, 'str'),

    casefold      = operation('casefold', None, 1, False, 'str'),
    cat           = operation('cat', None, 2, False, 'str'),
    decode        = operation('decode', None, 1, False, 'str'),
    extract       = operation('extract', None, 1, False, 'str'),
    extractall    = operation('extractall', None, 1, False, 'str'),
    get           = operation('get', None, 1, False, 'str'),
    get_dummies   = operation('get_dummies', None, 1, False, 'str'),
    index         = operation('index', None, 1, False, 'str'),
    join          = operation('join', None, 1, False, 'str'),
    normalize     = operation('normalize', None, 1, False, 'str'),
    partition     = operation('partition', None, 1, False, 'str'),
    repeat        = operation('repeat', None, 1, False, 'str'),
    rindex        = operation('rindex', None, 1, False, 'str'),
    rpartition    = operation('rpartition', None, 1, False, 'str'),
    translate     = operation('translate', None, 1, False, 'str'),
    zfill         = operation('zfill', None, 1, False, 'str'),
)


# Datetime --------------------------------------------------------------------

ops_dt = Namespace(
    ceil             = operation('ceil', 'elwise', 1, False, 'dt'),
    date             = operation('date', 'elwise', 1, True, 'dt'),
    day              = operation('day', 'elwise', 1, True, 'dt'),
    day_name         = operation('day_name', 'elwise', 1, False, 'dt'),
    dayofweek        = operation('dayofweek', 'elwise', 1, True, 'dt'),
    dayofyear        = operation('dayofyear', 'elwise', 1, True, 'dt'),
    days_in_month    = operation('days_in_month', 'elwise', 1, True, 'dt'),
    daysinmonth      = operation('daysinmonth', 'elwise', 1, True, 'dt'),
    floor            = operation('floor', 'elwise', 1, False, 'dt'),
    hour             = operation('hour', 'elwise', 1, True, 'dt'),
    is_leap_year     = operation('is_leap_year', 'elwise', 1, True, 'dt'),
    is_month_end     = operation('is_month_end', 'elwise', 1, True, 'dt'),
    is_month_start   = operation('is_month_start', 'elwise', 1, True, 'dt'),
    is_quarter_end   = operation('is_quarter_end', 'elwise', 1, True, 'dt'),
    is_quarter_start = operation('is_quarter_start', 'elwise', 1, True, 'dt'),
    is_year_end      = operation('is_year_end', 'elwise', 1, True, 'dt'),
    is_year_start    = operation('is_year_start', 'elwise', 1, True, 'dt'),
    microsecond      = operation('microsecond', 'elwise', 1, True, 'dt'),
    minute           = operation('minute', 'elwise', 1, True, 'dt'),
    month            = operation('month', 'elwise', 1, True, 'dt'),
    month_name       = operation('month_name', 'elwise', 1, False, 'dt'),
    nanosecond       = operation('nanosecond', 'elwise', 1, True, 'dt'),
    normalize        = operation('normalize', 'elwise', 1, False, 'dt'),
    quarter          = operation('quarter', 'elwise', 1, True, 'dt'),
    qyear            = operation('qyear', 'elwise', 1, True, 'dt'),
    round            = operation('round', 'elwise', 1, False, 'dt'),
    second           = operation('second', 'elwise', 1, True, 'dt'),
    strftime         = operation('strftime', 'elwise', 1, False, 'dt'),
    time             = operation('time', 'elwise', 1, True, 'dt'),
    timetz           = operation('timetz', 'elwise', 1, True, 'dt'),
    to_period        = operation('to_period', 'elwise', 1, False, 'dt'),
    to_pydatetime    = operation('to_pydatetime', 'elwise', 1, False, 'dt'),
    to_pytimedelta   = operation('to_pytimedelta', 'elwise', 1, True, 'dt'),
    to_timestamp     = operation('to_timestamp', 'elwise', 1, False, 'dt'),
    total_seconds    = operation('total_seconds', 'elwise', 1, True, 'dt'),
    tz_convert       = operation('tz_convert', 'elwise', 1, False, 'dt'),
    tz_localize      = operation('tz_localize', 'elwise', 1, False, 'dt'),
    week             = operation('week', 'elwise', 1, True, 'dt'),
    weekday          = operation('weekday', 'elwise', 1, True, 'dt'),
    weekofyear       = operation('weekofyear', 'elwise', 1, True, 'dt'),
    year             = operation('year', 'elwise', 1, True, 'dt'),

    freq = operation('freq', 'singleton', 1, True, 'dt'),
    tz   = operation('tz', 'singleton', 1, True, 'dt'),

    components = operation('components', None, 1, True, 'dt'),
    end_time   = operation('end_time', None, 1, True, 'dt'),
    start_time = operation('start_time', None, 1, True, 'dt'),
    )

# TODO: make aliases
#ops_dt_aliases = "days", "microseconds", "nanoseconds", "seconds"] 

ops_dt_plain = Namespace(
    # Note: only between_time is useful to siuba!
    asfreq            = operation('asfreq', 'elwise', 1),
    between_time      = operation('between_time', None, 1),

    asof              = operation('asof', 'window', 1),
    shift             = operation('shift', 'window', 1),
    tshift            = operation('tshift', None, 1),

    at_time           = operation('at_time', None, 1),
    first_valid_index = operation('first_valid_index', None, 1),
    last_valid_index  = operation('last_valid_index', None, 1),
    resample          = operation('resample', None, 1),
    slice_shift       = operation('slice_shift', None, 1),
    )

ops_missing = Namespace(
    fillna      = operation('fillna', 'elwise', 1),
    isna        = operation('isna', 'elwise', 1),
    isnull      = operation('isnull', 'elwise', 1),
    notna       = operation('notna', 'elwise', 1),
    notnull     = operation('notnull', 'elwise', 1),

    bfill       = operation('bfill', 'window', 1),
    ffill       = operation('ffill', 'window', 1),

    dropna      = operation('dropna', None, 1),
    interpolate = operation('interpolate', None, 1),
    )

# =============================================================================
# Not broadly supported 
# =============================================================================

# Function application --------------------------------------------------------

ops_apply = Namespace(
    agg       = operation('agg', None, 1, True, None),
    aggregate = operation('aggregate', None, 1, True, None),
    apply     = operation('apply', None, 1, True, None),
    ewm       = operation('ewm', None, 1, True, None),
    expanding = operation('expanding', None, 1, True, None),
    groupby   = operation('groupby', None, 1, True, None),
    map       = operation('map', None, 1, True, None),
    pipe      = operation('pipe', None, 1, True, None),
    rolling   = operation('rolling', None, 1, True, None),
    transform = operation('transform', None, 1, True, None),
    )

# Attributes ------------------------------------------------------------------

ops_attrs = Namespace(
    T            = operation('T', None, 1, True, None),
    array        = operation('array', None, 1, True, None),
    attrs        = operation('attrs', None, 1, True, None),
    axes         = operation('axes', None, 1, True, None),
    base         = operation('base', None, 1, True, None),
    data         = operation('data', None, 1, True, None),
    dtype        = operation('dtype', 'singleton', 1, True, None),
    dtypes       = operation('dtypes', None, 1, True, None),
    empty        = operation('empty', 'agg', 1, True, None),
    flags        = operation('flags', None, 1, True, None),
    ftype        = operation('ftype', None, 1, True, None),
    ftypes       = operation('ftypes', None, 1, True, None),
    hasnans      = operation('hasnans', 'agg', 1, True, None),
    index        = operation('index', None, 1, True, None),
    itemsize     = operation('itemsize', None, 1, True, None),
    memory_usage = operation('memory_usage', 'agg', 1, True, None),
    name         = operation('name', 'singleton', 1, True, None),
    nbytes       = operation('nbytes', 'agg', 1, True, None),
    ndim         = operation('ndim', 'singleton', 1, True, None),
    put          = operation('put', None, 1, True, None),
    shape        = operation('shape', None, 1, True, None),
    size         = operation('size', 'agg', 1, True, None),
    strides      = operation('strides', None, 1, True, None),
    transpose    = operation('transpose', None, 1),
    values       = operation('values', None, 1, True, None),
    )


# Categoricals ----------------------------------------------------------------

ops_cat = Namespace(
    add_categories = operation('add_categories', 'elwise', 1, False, 'cat'),
    as_ordered = operation('as_ordered', 'elwise', 1, False, 'cat'),
    as_unordered = operation('as_unordered', 'elwise', 1, False, 'cat'),
    categories = operation('categories', None, 1, True, 'cat'),
    codes = operation('codes', None, 1, True, 'cat'),
    ordered = operation('ordered', None, 1, True, 'cat'),
    remove_categories = operation('remove_categories', 'elwise', 1, False, 'cat'),
    remove_unused_categories = operation('remove_unused_categories', 'window', 1, False, 'cat'),
    rename_categories = operation('rename_categories', 'elwise', 1, False, 'cat'),
    reorder_categories = operation('reorder_categories', 'elwise', 1, False, 'cat'),
    set_categories = operation('set_categories', 'elwise', 1, False, 'cat'),
    )

# Combining ---
ops_combine = Namespace(
    combine       = operation('combine', None, 2),
    combine_first = operation('combine_first', None, 2),
    append        = operation('append', None, 2),
    replace       = operation('replace', 'elwise', 1),
    update        = operation('update', None, 1),
    )

ops_convert = Namespace(
    __array__ = operation('__array__', None, 1),
    astype = operation('astype', 'elwise', 1),
    bool = operation('bool', 'agg', 1, True, None),
    convert_dtypes = operation('convert_dtypes', 'window', 1),
    copy = operation('copy', 'elwise', 1),
    get_values = operation('get_values', None, 1),
    infer_objects = operation('infer_objects', None, 1),
    to_list = operation('to_list', None, 1),
    to_numpy = operation('to_numpy', None, 1),
    to_period = operation('to_period', 'elwise', 1),
    to_timestamp = operation('to_timestamp', None, 1),
    tolist = operation('tolist', None, 1),
    )

ops_index = Namespace(
    add_prefix = operation('add_prefix', None, 1),
    add_suffix = operation('add_suffix', None, 1),
    align = operation('align', None, 1, True, None),
    at = operation('at', 'window', 1),
    drop = operation('drop', None, 1, True, None),
    drop_duplicates = operation('drop_duplicates', None, 1),
    droplevel = operation('droplevel', None, 1, True, None),
    duplicated = operation('duplicated', 'window', 1),
    equals = operation('equals', 'agg', 2),
    filter = operation('filter', None, 1),
    first = operation('first', None, 1),
    get = operation('get', 'window', 1),
    head = operation('head', 'window', 1),
    iat = operation('iat', 'window', 1),
    idxmax = operation('idxmax', 'agg', 1),
    idxmin = operation('idxmin', 'agg', 1),
    iloc = operation('iloc', 'window', 1),
    item = operation('item', None, 1),
    items = operation('items', None, 1),
    iteritems = operation('iteritems', None, 1),
    last = operation('last', None, 1),
    loc = operation('loc', None, 1),
    mask = operation('mask', 'elwise', 2),
    pop = operation('pop', None, 1),
    reindex = operation('reindex', None, 1, True, None),
    reindex_like = operation('reindex_like', None, 1, True, None),
    rename = operation('rename', None, 1),
    rename_axis = operation('rename_axis', None, 1, True, None),
    reset_index = operation('reset_index', None, 1),
    sample = operation('sample', None, 1),
    set_axis = operation('set_axis', None, 1, True, None),
    tail = operation('tail', 'window', 1),
    take = operation('take', None, 1, True, None),
    truncate = operation('truncate', None, 1, True, None),
    where = operation('where', 'elwise', 2),
    xs = operation('xs', None, 1),
    )

ops_io = Namespace(
    to_clipboard = operation('to_clipboard', None, 1),
    to_csv = operation('to_csv', None, 1),
    to_dense = operation('to_dense', None, 1),
    to_dict = operation('to_dict', None, 1),
    to_excel = operation('to_excel', None, 1),
    to_frame = operation('to_frame', None, 1),
    to_hdf = operation('to_hdf', None, 1),
    to_json = operation('to_json', None, 1),
    to_latex = operation('to_latex', None, 1),
    to_markdown = operation('to_markdown', None, 1),
    to_msgpack = operation('to_msgpack', None, 1),
    to_pickle = operation('to_pickle', None, 1),
    to_sql = operation('to_sql', None, 1),
    to_string = operation('to_string', None, 1),
    to_xarray = operation('to_xarray', None, 1),
    )

ops_reshape = Namespace(
    argmax = operation('argmax', None, 1, True, None),
    argmin = operation('argmin', None, 1, True, None),
    argsort = operation('argsort', None, 1, True, None),
    explode = operation('explode', None, 1),
    ravel = operation('ravel', None, 1),
    reorder_levels = operation('reorder_levels', None, 1, True, None),
    repeat = operation('repeat', None, 1),
    searchsorted = operation('searchsorted', None, 1),
    sort_index = operation('sort_index', None, 1),
    sort_values = operation('sort_values', None, 1),
    squeeze = operation('squeeze', None, 1),
    swapaxes = operation('swapaxes', None, 1, True, None),
    swaplevel = operation('swaplevel', None, 1, True, None),
    unstack = operation('unstack', None, 1, True, None),
    view = operation('view', None, 1, True, None),
    )

ops_sparse = Namespace(
    density = operation('density', None, 1, True, 'sparse'),
    fill_value = operation('fill_value', None, 1, True, 'sparse'),
    from_coo = operation('from_coo', None, 1, False, 'sparse'),
    npoints = operation('npoints', None, 1, True, 'sparse'),
    sp_values = operation('sp_values', None, 1, True, 'sparse'),
    to_coo = operation('to_coo', None, 1, False, 'sparse'),
    to_dense = operation('to_dense', None, 1, False, 'sparse'),
    )


# keys method is necessary for Namespace to be a mapping, so we can dict unpack.
# For now, just put the keys generic function on its own. It's not a commonly
# used method, especially in siuba (since it just gets the index...)
keys = operation('keys', None, 1)

# Put all ops into a dict, that can be used by e.g. CallTreeLocal -------------

# This is a hacky way to get all the operations on a single namespace, since
# SimpleNamepace does not allow iteration
PLAIN_OPS = dict(
        # TODO: modify CallTreeLocal to take modules for accessors
        **ops_infix,
        **ops_infix_methods,
        **ops_apply,
        **ops_attrs,
        **ops_compute,
        **ops_dt_plain,
        **ops_missing,
        **ops_combine,
        **ops_convert,
        **ops_index,
        **ops_reshape,
        keys = keys,
        )

ACCESSOR_OPS = dict(
        **{'str.' + k: v for k,v in dict(ops_str).items()},
        **{'dt.' + k: v for k,v in dict(ops_dt).items()},
        **{'cat.' + k: v for k,v in dict(ops_cat).items()},
        **{'sparse.' + k: v for k,v in dict(ops_sparse).items()},
        )

ALL_OPS = {**PLAIN_OPS, **ACCESSOR_OPS}

ALL_ACCESSORS = set()
ALL_PROPERTIES = set()

for _k, _f in ALL_OPS.items():
    if _f.operation.is_property: ALL_PROPERTIES.add(_k)
    access = _f.operation.accessor
    if access is not None: ALL_ACCESSORS.add(access)




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-filter.R
"""
    
from siuba import _, filter, group_by, arrange, summarize
from siuba.dply.vector import row_number, desc
import pandas as pd

import pytest

from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql

DATA = pd.DataFrame({
    "x": [1,1,1,1],
    "y": [1,1,2,2],
    "z": [1,2,1,2]
    })


def test_filter_basic(backend):
    df = data_frame(x = [1,2,3,4,5], y = [5,4,3,2,1])
    dfs = backend.load_df(df)

    assert_equal_query(dfs, filter(_.x > 3), df[lambda _: _.x > 3])

def test_filter_basic_two_args(backend):
    df = data_frame(x = [1,2,3,4,5], y = [5,4,3,2,1])
    dfs = backend.load_df(df)

    assert_equal_query(dfs, filter(_.x > 3, _.y < 2), df[lambda _: (_.x > 3) & (_.y < 2)])

def test_filter_via_group_by(backend):
    df = data_frame(
            x = range(1, 11),
            g = [1]*5 + [2]*5
            )

    dfs = backend.load_df(df)

    assert_equal_query(
            dfs,
            # arrange is required to ensure order in sql dbs
            arrange(_.x) >> group_by(_.g) >> filter(row_number(_) < 3),
            data_frame(x = [1,2,6,7], g = [1,1,2,2])
            )


def test_filter_via_group_by_agg(backend):
    dfs = backend.load_df(x = range(1,11), g = [1]*5 + [2]*5)

    assert_equal_query(
            dfs,
            group_by(_.g) >> filter(_.x > _.x.mean()),
            data_frame(x = [4, 5, 9, 10], g = [1, 1, 2, 2])
            )


def test_filter_via_group_by_agg_two_args(backend):
    dfs = backend.load_df(x = range(1,11), g = [1]*5 + [2]*5)

    assert_equal_query(
            dfs,
            group_by(_.g) >> filter(_.x > _.x.mean(), _.x != _.x.max()),
            data_frame(x = [4, 9], g = [1, 2])
            )


@backend_sql("TODO: pandas - implement arrange over group by")
def test_filter_via_group_by_arrange(backend):
    dfs = backend.load_df(x = [3,2,1] + [2,3,4], g = [1]*3 + [2]*3)

    assert_equal_query(
            dfs,
            group_by(_.g) >> arrange(_.x) >> filter(_.x.cumsum() > 3),
            data_frame(x = [3, 3, 4], g = [1, 2, 2])
            )

@backend_sql("TODO: pandas - implement arrange over group by")
def test_filter_via_group_by_desc_arrange(backend):
    dfs = backend.load_df(x = [3,2,1] + [2,3,4], g = [1]*3 + [2]*3)

    assert_equal_query(
            dfs,
            group_by(_.g) >> arrange(desc(_.x)) >> filter(_.x.cumsum() > 3),
            data_frame(x = [2, 1, 4, 3, 2], g = [1, 1, 2, 2, 2])
            )

def test_filter_before_summarize(backend):
    dfs = backend.load_df(x = [1,2,3], g = ["a", "b", "b"])

    assert_equal_query(
            dfs,
            filter(_.x > 2) >> summarize(z=_.x.mean()),
            data_frame(z = [3]),
            # sql backends vary in the type .mean() returns
            check_dtype=False
            )

def test_filter_before_summarize_grouped(backend):
    dfs = backend.load_df(x = [1,2,3], g = ["a", "a", "b"])

    assert_equal_query(
            dfs,
            group_by(_.g) >> filter(_.x.mean() > 2) >> summarize(z=_.x.mean()),
            data_frame(g = ["b"], z = [3]),
            # sql backends vary in the type .mean() returns
            check_dtype=False
            )

